# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import os
import shutil

import numpy as np

from elektronn.net import netutils
import trainutils
import elektronn.examples


class MasterConfig(object):
    def __init__(self):
        """
        This class hard-codes the distribution-wide default values
        """
        ### Toolkit Setup ### ------------------------------------------------------------------------------------------------------------
        self.save_path = "~/CNN_Training/"  # (*) <String>: where to create the CNN directory. In this directory a new folder is created with the name of the model
        self.plot_on = True  # <Bool>: whether to create plots of the errors etc.
        self.print_status = True  # <Bool>: whether to print Training status to std.out
        self.device = False  # False (use .theanorc value) or int (use gpu<i>)
        self.param_save_h = 1.0  # hours: frequency to save a permanent parameter snapshot
        self.initial_prev_h = 1.0  # hours: time after which first preview is made
        self.prev_save_h = 3.0  # hours: frequency to create previews

        ### Paths and General ### ------------------------------------------------------------------------------------------------------------
        self.save_name = 'Test'  # (*) <String>: with the save_name.
        self.overwrite = True  # <Bool>: whether to delete/overwrite existing directory
        self.save_path += self.save_name + '/'
        self.param_file = None  # <String>/<None>: optional parameter file to initialise weights from

        ### Network Architecture ### ------------------------------------------------------------------------------------------------------------
        # Note: the last layer is added automatically (with n_lab outputs)
        self.activation_func = 'relu'  # <String> or <List> of <String>s: [relu], abs, linear, sig, tanh, (globally or per layer)
        self.batch_size = 1  # (*) <Int>: number of different slices/examples used for one optimisation step
        self.dropout_rates = []  # (*) <List> of <Float>>(0,1) or <Float>(0,1): "fail"-rates (globally or per layer)
        # The last layer never has dropout.
        # Conv layers
        self.n_dim = 2  # (*) 2/3:  for non image data (mode 'vect-scalar') this is ignored
        self.desired_input = 100  # (*) <Int> or 2/3-Tuple: in (x,y)/(x,y,z)-order for anisotropic CNN
        self.filters = []  # (*) <List> of <Int> or <List> of <2/3-tuples> or []: filter shapes in (x,y)/(x,y,z)-order
        self.pool = []  # (*) <List> of <Int> or <List> of <2/3-tuples> or []: pool shapes in (x,y)/(x,y,z)-order
        self.nof_filters = []  # (*) <List> of <Int> / []: number of feature maps per layer
        self.pooling_mode = 'max'  # ($) <String> or <List> of <String>: select pooling function (globally or per layer)
        # available: 'max', 'maxabs'
        self.MFP = []  # (*) <List> of <Int>{0,1}/False: whether to apply Max-Fragment-Pooling (globally or per layer)

        # Other
        self.rnn_layer_kwargs = None  # ($) <Dict>/<None>: if in use: dict(n_hid=Int,  activation_func='tanh', iterations=None/Int)
        self.MLP_layers = []  # (*) <List> of <Int>: numbers of filters for fully connected layers (after conv layers)
        self.target = 'nll'  # <String>: 'nll' or 'regression'

        ### Data CNN ### (the whole block is ignored for mode 'vect-scalar') ### ----------------------------------------------------------------
        self.data_path = '/'  # (*) <String>: Path to data dir
        self.label_path = '/'  # (*) <String>: Path to label dir
        self.d_files = []  # (*) <List> of tuples: (file name, key of h5 data set)
        self.l_files = []  # (*) <List> of tuples: (file name, key of h5 data set)
        self.cube_prios = None  # <List>/<None>: sampling priorities for cubes or None, then: sampling ~ example_size
        # will be normalised internally
        self.valid_cubes = []  # <List>: of cube indices (from the file-lists) to use as validation data,
        # may be empty
        self.example_ignore_threshold = 0.0  # <Float>: If the fraction of negative labels in an example patch exceeds
        # this threshold this example is discarded
        self.grey_augment_channels = [0]  # (*) <List>:  of integer channel-indices to apply grey augmentation, use [] to disable.
        # It distorts the histogram of the raw images (darker, lighter, more/less contrast)
        self.use_example_weights = False  # ($) <Bool>: Whether to use weights for the examples (e.g. for Boosting-like training)
        self.flip_data = True  # <Bool>: whether to flip/rotate/mirror data randomly (augmentation)
        self.anisotropic_data = True  # (*) <Bool>: if True 2D slices are only cut in z-direction, otherwise all 3 alignments are used
        self.lazy_labels = False  # ($) <Bool>: Special Training with lazy annotations
        self.warp_on = False  # <Bool>/<Float>(0,1): Warping augmentations (CPU-intense, use background processes!)
        # If <Float>: warping is applied to this fraction of examples e.g. 0.5 --> every 2nd example
        self.border_mode = "crop"  # <String>: Only applicable for *img-scalar*. If the CNN does not allow the original size of
        # the images the following options are available:
        # "crop"  : if img too big, cut the images to the next smaller valid input size,
        # "keep"  : if img too big, keep the size and cut smaller patches with translations                               
        # "0-pad" : if img to small, padd to the next bigger valid input with zeros,
        # "mirror": if img to small, padd to the next bigger valid input by mirroring along border
        # "reject": if img size too small/big, throw exception
        self.pre_process = None  #  "standardise"/None: (0-mean, 1-std) (over all pixels)
        self.zchxy_order = False  # <Bool>: set to True if data is in (z, (ch,) x, y) order, otherwise (ch, x, y, z) is assume
        self.upright_x = False  # <Bool>: If true, mirroring is only applied horizontally (e.g. for outdoor images or handwriting)
        self.downsample_xy = False  # <Bool>/<Int> downsample by this factor, or not at all if False

        ### Data Preview ### (only for img-img) -------------------------------------------------------------------------------------------------
        self.preview_data_path = None  # <String>/<None>: path to a h5-file that contains data to make preview predictions
        # it must contain a list of image cubes (normalised between 0 and 255) in the shape ((ch,) x,y,z)
        self.preview_kwargs = dict(export_class=1, max_z_pred=5)  # <Dict>: specification of preview to create

        ### Data Common ### ---------------------------------------------------------------------------------------------------------------------
        self.mode = 'img-img'  # (*) <String>: combination of data and label types: 'img-img', 'img-scalar', vect-scalar'
        self.n_lab = None  # (*) <Int>/<None>: (None means auto detect, very slow, don't do this!)
        self.background_processes = False  # <Bool>/<Int>: whether to "pre-fetch" batches in separate background
        # process, <Bool> or number of processes (True-->2)

        ### Data Alternative / vect-scalar ### (this may replace the above CNN block) ### -------------------------------------------------------
        self.data_class_name = None  # <String>: Name of Data Class in TrainData
        self.data_load_kwargs = dict()  # <Dict>: Arguments to init Data Class
        self.data_batch_kwargs = dict()  # <Dict>: Arguments for getbach method of Data Class (for training set only!)
        # The batch_size argument is added internally and needn't be specified here

        ### Optimisation Options ### ------------------------------------------------------------------------------------------------------------
        self.n_steps = 10**12  # (*) <Int>: number of update steps
        self.max_runtime = 24 * 3600  # (*) <Int>: maximal Training time in seconds (overrides n_steps)
        self.history_freq = [500]  # (*) <List> of single <Int>: create plots, print status and test model after x steps
        self.monitor_batch_size = 10  # (*) <Int>: number of patches to test model on (valid and train subset)

        self.weight_decay = False  # ($) False/<Float>: weighting of L2-norm on weights ("lambda"), False is equal to 0.0
        self.class_weights = None  # ($) <None> or <List> of <Float>: weights for the classes, will be normalised internally
        self.label_prop_thresh = None  # ($) <None>/<Float>: if <Float> value is set, unlabelled examples (-1) will be trained on the
        # current prediction if the predicted probability exceeds this threshold.

        self.optimizer = 'SGD'  # ($) <String>: [SGD]/CG/RPORP/LBFGS method for training

        self.LR_decay = 0.993  # (*) <Float>: decay of SGD learning rate w.r.t to an interval of 1000 update steps
        self.LR_schedule = None  # (*) <List> of tuples/<None>: (#iteration, new_LR), if used this sets the LR
        # at the specified iteration steps to the specified value. This is independent of the decay.

        # ---------------------------------------------------------------------------------------------------------------------------------------
        # SGD
        self.SGD_params = dict(LR=0.001,
                               momentum=0.9)  # (*) <Dict>: initial learning rate and momentum

        self.SSGD_params = dict(LR=0.01,
                                momentum=0.9,
                                var_momentum=0.9,
                                var_cuttoff=3.0)  # <Dict>

        # RPROP ($)
        self.RPROP_params = dict(penalty=0.35,
                                 gain=0.2,
                                 beta=0.7,
                                 initial_update_size=1e-4)  # <Dict>

        # Adam ($)
        self.Adam_params = dict(LR=0.001,
                                beta1=0.9,
                                beta2=0.999,
                                epsilon=1e-8)  # <Dict>

        # CG ($)
        self.CG_params       = dict(n_steps=4,       # update steps per same batch 3 <--> 6
                               alpha=0.35,           # termination criterion of line search, must be <= 0.35
                               beta=0.75,            # precision of line search,  imprecise 0.5 <--> 0.9 precise
                               max_step=0.02,        # similar to learning rate in SGD 0.1 <--> 0.001.
                               min_step=8e-5)

        # LBFGS ($)
        self.LBFGS_params    = dict(maxfun= 40,      # function evaluations
                               maxiter= 4,           # iterations
                               m= 10,                # maximum number of variable metric corrections
                               factr= 1e2,           # factor of machine precision as termination criterion
                               pgtol= 1e-9,          # projected gradient tolerance
                               iprint= -1)           # set to 0 for direct printing of steps
        # ---------------------------------------------------------------------------------------------------------------------------------------
        self.__doc__ = ""  # Just a hack


class DefaultConfig(MasterConfig):
    def __init__(self):
        """
        This class overwrites the master values with values from a user file (if present)
        """
        super(DefaultConfig,
              self).__init__()  # gives initiall values for all attributes

        config_dict = {}
        user_path = os.path.expanduser('~/.elektronn.config')
        if not os.path.exists(user_path):
            pass  # Maybe message that user should create own file
        else:
            try:
                print "Reading User Default Config"
                execfile(user_path, {}, config_dict)
            except Exception, e:
                raise RuntimeError("The user config file %s does exist, but an error happend during reading, it might contain invalid code. Error: \n  %s"
                                   %(user_path, e))

            for key in config_dict:
                setattr(self, key, config_dict[key])


default_config = DefaultConfig()

### Configuration ############################################################################################


class Config(object):
    """
    Configuration object to manage the parameters of ``trainingInstance``

    The ``monitor_batch_size`` is automatically fixed to a multiple of the ``batch_size``.
    An attribute ``dimensions`` of type ``Net.netutils.CNNCalcquotesulator`` is created that checks if the CNN
    architecture (combination of filter sizes and poolings) is valid and determines the input shape closest
    to the ``desired_input``
    The top level script (``trainer_file``), the config, the ``Net`` module and the ``Training`` module
    are backed up into the CNN directory automatically. The Backup is intended to contain all code to
    reproduce the CNN Training


    Parameters
    ----------

    config_file: string
      Path to a CNN config file
    gpu: int
      Specifying id of GPU to initialise for usage. E.g. 1 --> "gpu1", None will initialise gpu0,\
      False will not initialise any GPU. This only works if "device" is not set in ``.theanorc`` or if theano
      has not been imported up to now. If the initialisation fails an error will be printed but the script
      will not crash.
    trainer_file: string
      Path to the ``NetTrainer``-script (or any other top level script that drives the Training).\
      The path is needed to backup the script in the CNN directory
    use_existing_dir: Bool
      Do not create a new directory for the CNN if True
    override_MFP_to_active: Bool
      If true, activates MFP in all layers where possible, ignoring the configuration in the config file.
      This is useful for prediction using a config file from training. (only for CNN)
    override_input_size_with: tuple or None
      Similar as above, this can be used to impose another input size than specified in the config file. (only for CNN)
    """

    mandatory_vars = ['SGD_params',
                      'batch_size',
                      'max_runtime',
                      'monitor_batch_size',
                      'n_steps',
                      'n_lab',
                      'save_name',
                      'mode', ]

    mandatory_data = ['d_files', 'data_path', 'l_files', 'label_path', ]

    mandatory_cnn = ['desired_input', 'filters', 'n_dim', 'nof_filters', 'pool'
                     ]

    mandatory_mlp = ['MLP_layers']

    def __init__(self,
                 config_file,
                 gpu,
                 trainer_file,
                 use_existing_dir=False,
                 override_MFP_to_active=False,
                 imposed_input_size=None):

        super(Config, self).__init__()

        self.imposed_input_size = imposed_input_size
        self.override_MFP_to_active = override_MFP_to_active

        # read and process the config parameters
        self.default_config = default_config
        self.config_file = os.path.expanduser(config_file)

        # If file is not found, look if it's an example file:
        if not os.path.isfile(self.config_file):
            example_path = os.path.join(os.path.dirname(elektronn.examples.__file__), config_file)
            if os.path.isfile(example_path):
                self.config_file = example_path
            else:
                raise RuntimeError('Config file %s could not be found.' % config_file)

        custom_dict = self.parseConfig()
        self.fixValues()
        self.typeChecks()
        self.checkConfig(custom_dict)

        # for image data handle the architecutre / input sizes
        if self.mode != 'vect-scalar':
            force_center = True if self.mode == 'img-img' else False
            self.dimensions = netutils.CNNCalculator(self.filters,
                                                     self.pool,
                                                     self.desired_input,
                                                     MFP=self.MFP,
                                                     force_center=force_center,
                                                     n_dim=self.n_dim)

            if self.mode == 'img-scalar':
                if np.any(np.not_equal(self.dimensions.input,
                                       self.desired_input)):
                    if self.border_mode in ['crop', 'keep']:
                        pass  # the next smaller input is selected anyway
                    elif self.border_mode in ['mirror', '0-pad']:
                        # select the next bigger input size
                        self.dimensions.input = map(lambda x: x[0] + x[1], zip(
                            self.dimensions.input,
                            self.dimensions.pred_stride))
                    else:  # 'reject'
                        raise ValueError(
                            "The CNN architecture does not allow the input size of the images and cropping/\
padding was disabled. Use an architecture that has matching input size, change the image sizes \
or use a different boarder mode.")

            print "Selected patch-size for CNN input:", self.dimensions
            self.patch_size = self.dimensions.input
            if not hasattr(self.patch_size, '__len__'):
                self.patch_size = (self.patch_size, ) * self.n_dim

        if not use_existing_dir:
            self.backupScripts(trainer_file)

        trainutils.initGPU(gpu)

    def parseConfig(self):
        print "Reading config-file %s" % (self.config_file, )
        config_dict = default_config.__dict__  # take attributes from default config
        self._allowed = config_dict.keys()
        custom_dict = {}
        execfile(self.config_file, {}, custom_dict)
        config_dict.update(custom_dict)

        for key in config_dict:
            if key in self._allowed:
                if key in ['save_path', 'param_file', 'data_path',
                           'label_path', 'preview_data_path', 'data_class_name']:
                    try:
                        if key == 'data_class_name' and isinstance(config_dict[key], tuple):
                            config_dict[key][0] = os.path.expanduser(config_dict[key][0])
                        else:
                            config_dict[key] = os.path.expanduser(config_dict[key])
                    except:
                        pass

                setattr(self, key, config_dict[key])
            else:
                min_dist = np.argmin([self.levenshtein(key, x) for x in self._allowed])
                raise ValueError("<" + str(key) + ">  is not valid configuration variable.\n\
          Did you mean <" + str(self._allowed[min_dist]) + ">?")

        self.n_layers = len(self.filters) + len(self.MLP_layers)  # rnn layer is not counted here
        return custom_dict

    def fixValues(self):
        if self.monitor_batch_size % self.batch_size != 0:  # make the monitor size a multiple of the normal
            self.monitor_batch_size = max(self.batch_size,
                                          (self.monitor_batch_size - self.monitor_batch_size % self.batch_size))

        if self.n_dim == 3:  # The user gives the shapes in xyz but internal functions require zxy
            self.filters = trainutils.xyz2zyx(self.filters)
            self.pool = trainutils.xyz2zyx(self.pool)
            if hasattr(self.desired_input, '__len__'):  # also swap this
                self.desired_input = self.desired_input[2:3] + self.desired_input[0:2]

        # Fix various configuration parameter so standard format
        if (self.MFP == False) or (self.MFP is None) or (self.MFP == []):
            self.MFP = [0, ] * len(self.filters)
        elif self.MFP == True:
            self.MFP = [1, ] * len(self.filters)

        if self.override_MFP_to_active:
            self.MFP = list(np.greater(map(np.max, self.pool), 1))  # activate in all layers that have a pool factor > 1
            self.batch_size = 1

        if self.imposed_input_size is not None:
            self.desired_input = self.imposed_input_size

        if not (isinstance(self.pooling_mode, list) or isinstance(self.pooling_mode, tuple)):
            self.pooling_mode = [self.pooling_mode, ] * len(self.filters)

        if not (isinstance(self.activation_func, list) or isinstance(self.activation_func, tuple)):
            self.activation_func = [self.activation_func, ] * self.n_layers

        if not hasattr(self.dropout_rates, '__len__'):  # may also be None --> list of Nones
            self.dropout_rates = (self.dropout_rates, ) * self.n_layers

        if self.class_weights is not None:
            self.class_weights = np.array(self.class_weights, dtype=np.float32)
            #self.class_weights /= np.sum(self.class_weights) # normalise weights
            print "WARNING: weight normalisation is suspendend but this contradicts doc!"

        if self.example_ignore_threshold is None:
            self.example_ignore_threshold = 0.0

    def typeChecks(self):
        is_string = lambda x: isinstance(x, str) or (x is None)
        is_int = lambda x: isinstance(x, bool) or isinstance(x, int)
        is_list = lambda x: isinstance(x, list) or isinstance(x, tuple) or (x is None) or is_int(x)
        is_dict = lambda x: isinstance(x, dict) or (x is None)
        is_float = lambda x: isinstance(x, float) or (x is None)

        for param in ['save_path', 'save_name', 'data_path', 'label_path',
                      'mode', 'border_mode', 'target', 'optimizer',
                      'param_file', 'preview_data_path']:
            assert is_string(getattr(self, param)), "Parameter %s must be a string or None" % (param)

        for param in ['overwrite', 'plot_on', 'print_status', 'flip_data',
                      'anisotropic_data', 'lazy_labels', 'use_example_weights',
                      'zchxy_order', 'upright_x', 'background_processes',
                      'batch_size', 'monitor_batch_size', 'n_dim']:
            assert is_int(getattr(self, param)), "Parameter %s must be a bool or an int" % (param)

        for param in ['d_files',
                      'l_files',
                      'valid_cubes',
                      'grey_augment_channels',
                      'dropout_rates',
                      'filters',
                      'pool',
                      'nof_filters',
                      'MFP',
                      'MLP_layers',
                      'history_freq',
                      'cube_prios',
                      'dropout_rates',
                      'LR_schedule', ]:
            assert is_list(getattr(self, param)), "Parameter %s must be a list or None" % (param)

        for param in ['data_load_kwargs', 'data_batch_kwargs',
                      'preview_kwargs', 'SGD_params', 'RPROP_params',
                      'CG_params', 'LBFGS_params', 'rnn_layer_kwargs']:
            assert is_dict(getattr(self, param)), "Parameter %s must be a dict or None" % (param)

        for param in ['example_ignore_threshold', ]:
            assert is_float(getattr(self, param)), "Parameter %s must be a Float" % (param)

    def checkConfig(self, custom_dict):
        mandatory = self.mandatory_vars
        if self.mode == 'img-img' or self.mode == 'img-scalar':
            mandatory.extend(self.mandatory_cnn)
        if self.mode == 'vect-scalar':
            mandatory.extend(self.mandatory_mlp)
        if self.data_class_name is None:
            mandatory.extend(self.mandatory_data)
        undefined_vars = []
        for var_name in mandatory:
            if not custom_dict.has_key(var_name):
                undefined_vars.append(var_name)

        if len(undefined_vars) > 0:
            print "\nWARNING: the following important CNN config variables were not found in the configuration file. The MASTER configuration value is in their place but further execution might fail. Undefined variables:\n%s\n" % (undefined_vars)

        self.n_layers = len(self.filters) + len(self.MLP_layers)
        assert self.n_layers==len(self.dropout_rates) or len(self.dropout_rates) == 0, "If dropout_rates is a list, there must be a dropout rate for every layer or it must be an empty list to disable dropout"
        assert self.lazy_labels and self.mode=='img-img' or not self.lazy_labels, "Lazy labels is only possible 'img-img' training"
        assert self.mode=='img-img' and len(self.MLP_layers)==0 or self.mode!='img-img', "'img-img' training does not allow MLP layers"
        assert self.mode=='img-scalar' and len(self.MLP_layers)!=0 or self.mode!='img-scalar', "'img-scalar' training requires MLP layers"
        assert self.mode=='vect-scalar' and len(self.filters)==0 or self.mode!='vect-scalar', "No ConvLayers allowed for 'vect-scalar' training"
        assert len(self.filters) == len(self.pool) == len(self.nof_filters), "All config lists for conv layers must have the same lenght"
        assert self.mode in ['img-img', 'img-scalar', 'vect-scalar'], "Unknown mode"
        assert (self.rnn_layer_kwargs is None and self.mode in ['img-img', 'img-scalar']) or self.mode=='vect-scalar', "RNN layers can only be used in 'vect-scalar' mode"
        targets = ['nll', 'regression', 'nll_mutiple_binary', 'nll_weak', 'affinity', 'malis']
        assert self.target in targets, "Invalid target functions, must be in %s" % (targets)

    def backupScripts(self, trainer_file):
        """
        Saves all python files into the folder specified by ``self.save_path``
        Also changes working directory to the ``save_path`` directory
        """
        if os.path.exists(self.save_path):
            if self.overwrite:
                print "Overwriting existing save directory: %s" % self.save_path
                assert self.save_path != './', ('Cannot delete current directory')
                shutil.rmtree(self.save_path)
            else:
                raise RuntimeError('The save directory does already exist!')

        os.makedirs(self.save_path)
        os.mkdir(self.save_path + 'Backup/')

        shutil.copy(trainer_file, self.save_path + "Backup/elektronn-train")
        os.chmod(self.save_path + "Backup/elektronn-train", 0755)

        #        import training
        #        trainer_dir = os.path.abspath(training.__file__)
        #        trainer_dir = '/'.join(trainer_dir.split('/')[:-1])+'/'
        #        shutil.copytree(trainer_dir, self.save_path+'Backup/training/')
        #
        #        import net
        #        net_dir = os.path.abspath(net.__file__)
        #        net_dir = '/'.join(net_dir.split('/')[:-1])+'/'
        #        shutil.copytree(net_dir, self.save_path+'Backup/net/')

        if self.config_file is not None:
            shutil.copy(self.config_file, self.save_path + "Backup/config.py")
            os.chmod(self.save_path + "Backup/config.py", 0755)

    def levenshtein(self, s1, s2):
        """
        Computes Levenshtein-distance between ``s1`` and ``s2`` strings
        Taken from: http://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
        """
        if len(s1) < len(s2):
            return self.levenshtein(s2, s1)
        # len(s1) >= len(s2)
        if len(s2) == 0:
            return len(s1)
        previous_row = range(len(s2) + 1)
        for i, c1 in enumerate(s1):
            current_row = [i + 1]
            for j, c2 in enumerate(s2):
                insertions = previous_row[j + 1] + 1  # j+1 instead of j since previous_row and current_row are one character longer
                deletions = current_row[j] + 1  # than s2
                substitutions = previous_row[j] + (c1 != c2)
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row
        return previous_row[-1]


if __name__ == "__main__":
    conf = Config('/docs/devel/ELEKTRONN/config_files/I-z.py', 0, '/docs/devel/ELEKTRONN/config_files/I-z.py')



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import time
import numpy as np
import theano
import theano.tensor as T
from theano.tensor.nnet import conv

import pooling
from netutils import initWeights
from gaborfilters import makeGaborFilters, blob


def getOutputShape(insh, fsh, pool, mfp, r=1):
    """
    Returns shape of convolution result from (bs, ch, x, y) * (nof, ch, xf, yf)
    """
    # insh = (1 if input_shape[0]==None else input_shape[0],)+input_shape[1:]

    if insh[0] is None:
        bs = None
    else:
        bs = insh[0] * np.prod(pool) if mfp else insh[0]
    ch = fsh[0] if r == 1 else fsh[0] // r
    x = (insh[2] - fsh[2] + 1) // pool[0]
    y = (insh[3] - fsh[3] + 1) // pool[1]
    return (bs, ch, x, y)


def getProbShape(output_shape, mfp_strides):
    """
    Given outputshape (bs, ch, x, y) and mfp_stride (sx, sy) returns shape of Class Prob output
    """
    if mfp_strides is None:
        return output_shape
    else:
        return (1, output_shape[1], output_shape[2] * mfp_strides[1], output_shape[3] * mfp_strides[1])


class ConvLayer2d(object):
    """
    Conv-Pool Layer of a CNN

    :type input: theano.tensor.dtensor4 ('batch', 'channel', x, y)
    :param input: symbolic image tensor, of shape input_shape

    :type input_shape: tuple or list of length 4
    :param input_shape: (batch size, num input feature maps, image height, image width)

    :type filter_shape: tuple or list of length 4
    :param filter_shape: (number of filters, input_channels, filter height,filter width)


    :type pool: int 2-tuple
    :param pool: the down-sampling (max-pooling) factor

    :type activation_func: string
    :param activation_func: Options: tanh, relu, sig, abs, linear, maxout <i>

    :type enable_dropout: Bool
    :param enable_dropout: whether to enable dropout in this layer. The default rate is 0.5 but it can be changed with
               self.activation_noise.set_value(set_value(np.float32(p)) or using cnn.setDropoutRates

    :type use_fragment_pooling: Bool
    :param use_fragment_pooling: whether to use max fragment pooling in this layer (MFP)

    :type reshape_output: Bool
    :param reshape_output: whether to reshape class_probabilities to (bs, cls, x, y) and re-assemble fragments
                         to dense images if MFP was enabled. Use this in for the last layer.

    :type mfp_offsets: list of list of ints
    :param mfp_offsets: this lists specifies the offsets that the MFP-fragments have w.r.t to the original patch.
                      Only needed if MFP is enabled.

    :type mfp_strides: list of int
    :param mfp_strides: the strides of the output in each dimension

    :type input_layer: layer object
    :param input_layer: just for keeping track of un-usual input layers

    :type W: np.ndarray or T.TensorVariable
    :param W: weight matrix. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)

    :type b: np.ndarray or T.TensorVariable
    :param b: bias vector. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)

    :type pooling_mode: str
    :param pooling_mode: 'max' or 'maxabs' where the first is normal maxpooling and the second also retains
                       sign of large negative values

    """

    def __init__(self,
                 input,
                 input_shape,
                 filter_shape,
                 pool,
                 activation_func,
                 enable_dropout,
                 use_fragment_pooling,
                 reshape_output,
                 mfp_offsets,
                 mfp_strides,
                 input_layer=None,
                 W=None,
                 b=None,
                 pooling_mode='max'):
        assert input_shape[1] == filter_shape[1]

        self.input = input
        self.pool = pool
        self.number_of_filters = filter_shape[0]
        self.filter_shape = filter_shape
        self.activation_func = activation_func
        self.input_shape = input_shape
        self.input_layer = input_layer
        self.mfp_strides = mfp_strides
        self.mfp_offsets = mfp_offsets
        self.reshape_output = reshape_output

        print "2DConv: input=", input_shape, "\tfilter=", filter_shape  #,"@std=",W_bound
        if W is None:
            W_values = np.asarray(
                initWeights(filter_shape,
                            scale='glorot',
                            mode='normal',
                            pool=pool),
                dtype=theano.config.floatX)
            self.W = theano.shared(W_values, name='W_conv', borrow=True)
        else:
            if isinstance(W, np.ndarray):
                self.W = theano.shared(
                    W.astype(np.float32),
                    name='W_conv',
                    borrow=True)
            else:
                assert isinstance(W, T.TensorVariable), "W must be either np.ndarray or theano var"
                self.W = W

        if b is None:
            n_out = filter_shape[0]
            if activation_func == 'relu' or activation_func == 'ReLU':
                norm = filter_shape[2] * filter_shape[3]
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1.0 / norm,
                        mode='const'),
                    dtype=theano.config.floatX)
            elif activation_func == 'sigmoid':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=0.5,
                        mode='const'),
                    dtype=theano.config.floatX)
            else:  # activation_func=='tanh':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1e-6,
                        mode='fix-uni'),
                    dtype=theano.config.floatX)
            self.b = theano.shared(value=b_values, borrow=True, name='b_conv')
        else:
            if isinstance(b, np.ndarray):
                self.b = theano.shared(
                    b.astype(np.float32),
                    name='b_conv',
                    borrow=True)
            else:
                assert isinstance(b, T.TensorVariable), "b must be either np.ndarray or theano var"
                self.b = b

        # store parameters of this layer
        self.params = [self.W, self.b]

        # convolve input feature maps with filters
        # shape of pooled_out , e.g.: (1,2,27,27) for 2 class-output
        self.conv_out = conv.conv2d(
                input=input,
                filters=self.W,
                border_mode='valid',
                filter_shape=filter_shape
        )  # image_shape = input_shape if input_shape[0] is not None else None)

        # down-sample each feature map individually, using maxpooling
        if np.any(pool != 1):
            pool_func = lambda x: pooling.pooling2d(x, pool_shape=pool, mode=pooling_mode)
            if use_fragment_pooling:
                pooled_out, self.mfp_offsets, self.mfp_strides = self.fragmentpool(
                    self.conv_out, pool, mfp_offsets, mfp_strides, pool_func)
            else:
                #pooled_out = downsample.max_pool_2d(input=self.conv_out, ds=pool, ignore_border=True)
                pooled_out = pool_func(self.conv_out)
        else:
            pooled_out = self.conv_out

        if enable_dropout:
            #print "Dropout: ACTIVE"
            self.activation_noise = theano.shared(np.float32(0.5), name='Dropout Rate')
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p = 1 - self.activation_noise
            self.dropout_gate = 1.0 / p * rng.binomial(
                (pooled_out.shape[2], pooled_out.shape[3]),
                1,
                p,
                dtype=theano.config.floatX)
            pooled_out = pooled_out * self.dropout_gate.dimshuffle(('x', 'x', 0, 1))

        # add the bias term. Since the bias is a vector (1D array), we first
        # reshape it to a tensor of shape (1,n_filters,1,1). Each bias will
        # thus be broadcasted across mini-batches and feature maps
        # width & height
        r = 1  # maxout factor
        lin_output = pooled_out + self.b.dimshuffle('x', 0, 'x', 'x')
        if activation_func == 'tanh':
            self.output = T.tanh(lin_output)  # shape: (batch_size, num_outputs)
        elif activation_func in ['ReLU', 'relu']:  #rectified linear unit
            self.output = lin_output * (lin_output > 0)  # shape: (batch_size, num_outputs)
        elif activation_func in ['sigmoid']:
            self.output = T.nnet.sigmoid(lin_output)
        elif activation_func in ['abs']:
            self.output = T.abs_(lin_output)
        elif activation_func in ['linear']:
            self.output = lin_output
        elif activation_func.startswith("maxout"):
            r = int(activation_func.split(" ")[1])
            assert r >= 2
            self.output = pooling.maxout(lin_output, factor=r)
        else:
            raise NotImplementedError()

        output_shape = getOutputShape(input_shape, filter_shape, pool, use_fragment_pooling, r)

        print "Output =", output_shape, "Dropout", ("ON," if enable_dropout else "OFF,"), "Act:",\
            activation_func, "pool:", pooling_mode
        self.output_shape = output_shape  # e.g. (None, 16, 100, 100)

        sh = lin_output.shape  # (bs,ch,x,y) # use this shape to reshape the output to image-shape after softmax
        sh = (sh[1], sh[2], sh[3], sh[0])  #(ch, x, y, bs)
        # put batchsize, at back --> (ch,x,y,bs), flatten this --> (ch, x*y*bs), swap labels --> (x*y*bs, ch)
        self.class_probabilities = T.nnet.softmax(lin_output.dimshuffle((1, 2, 3, 0)).flatten(2).dimshuffle((1, 0)))
        if reshape_output:
            self.reshapeoutput(sh)
            if use_fragment_pooling:
                self.fragmentstodense(sh)

            self.prob_shape = getProbShape(output_shape, self.mfp_strides)
            print "Class Prob Output =", self.prob_shape

        # compute prediction as class whose "probability" is maximal in symbolic form
        self.class_prediction = T.argmax(self.class_probabilities, axis=1)
        #output has shape e.g. (1,2,57,57); only the last two may change, 2 classes are predicted

    def fragmentpool(self, conv_out, pool, offsets, strides, pool_func):
        result = []
        offsets_new = []
        if offsets is None:
            offsets = np.array([[0, 0]])
        if strides is None:
            strides = [1, 1]

        sh = conv_out.shape

        for px in xrange(pool[0]):
            for py in xrange(pool[1]):
                pxy = pool_func(conv_out[:, :, px:sh[2] - pool[0] + px + 1, py:sh[3] - pool[1] + py + 1])
                result.append(pxy)
                for p in offsets:
                    new = p.copy()
                    new[0] += px * strides[0]
                    new[1] += py * strides[1]
                    offsets_new.append(new)

        result = T.concatenate(result, axis=0)
        offsets_new = np.array(offsets_new)
        strides = np.multiply(pool, strides)

        return result, offsets_new, strides

    def reshapeoutput(self, sh):
        # same shape as result, no significant time cost
        self.class_probabilities = self.class_probabilities.dimshuffle((1, 0)).reshape(sh).dimshuffle((3, 0, 1, 2))

    def fragmentstodense(self, sh):
        mfp_strides = self.mfp_strides
        example_stride = np.prod(mfp_strides)  # This stride is conceptually unneeded but theano-grad fails otherwise
        zero = np.array((0), dtype=theano.config.floatX)
        embedding = T.alloc(zero, 1, sh[0], sh[1] * mfp_strides[0], sh[2] * mfp_strides[1])
        ix = self.mfp_offsets
        for i, (n, m) in enumerate(ix):
            embedding = T.set_subtensor(
                embedding[:, :, n::mfp_strides[0], m::mfp_strides[1]], self.class_probabilities[i::example_stride])

        self.class_probabilities = embedding

    def randomizeWeights(self, scale='glorot', mode='uni'):
        n_out = self.filter_shape[0]
        if self.activation_func == 'relu':
            norm = self.filter_shape[2] * self.filter_shape[3]
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=1.0 / norm,
                    mode='const'),
                dtype=theano.config.floatX)
        elif self.activation_func == 'sigmoid':
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=0.5,
                    mode='const'),
                dtype=theano.config.floatX)
        else:  #self.activation_func=='tanh':
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=1e-6,
                    mode='fix-uni'),
                dtype=theano.config.floatX)

        W_values = np.asarray(
            initWeights(self.filter_shape,
                        scale,
                        mode,
                        pool=self.pool),
            dtype=theano.config.floatX)

        self.W.set_value(W_values)
        self.b.set_value(b_values)

    def gaborInitialisation(self):
        nof, channel, x, y = self.filter_shape
        if nof % 2 == 1:
            ga = makeGaborFilters(x,
                                  (nof - 1) / 2) / 10  # generate gabor filters
            ga = np.concatenate(
                (ga, blob(x)[:np.newaxis, :, :]))  # add a blob filter
        else:
            ga = makeGaborFilters(x,
                                  (nof - 2) / 2) / 10  # generate gabor filters
            noise = np.random.normal(loc=0, scale=0.02, size=(x, y))
            # add a blob and random filter
            ga = np.concatenate(
                (ga, blob(x)[np.newaxis, :, :] / 10, noise[np.newaxis, :, :]),
                axis=0)

        filters = np.zeros(self.filter_shape)
        for i in xrange(3):  # replicate filters for all colour channels
            filters[:, i, :, :] = ga

        self.W.set_value(np.asarray(filters, dtype=theano.config.floatX))

#  def negative_log_likelihood(self, y):
#    """
#    Returns the mean and instance-wise negative log-likelihood of the prediction
#    of this model under a given target distribution.
#
#    :type y: theano.tensor.TensorType
#    :param y: corresponds to a vector that gives for each example the
#              correct label
#
#    Note: we use the mean instead of the sum so that
#    the learning rate is less dependent on the batch size
#    """
#    # y.shape[0] is (symbolically) the number of rows in y, i.e.,
#    # number of examples (call it n) in the minibatch
#    # T.arange(y.shape[0]) is a symbolic vector which will contain [0,1,2,... n-1]
#    # T.log(self.p_y_given_x) is a matrix of
#    # Log-Probabilities (call it LP) with one row per example and
#    # one column per class LP[T.arange(y.shape[0]),y] is a vector
#    # v containing [LP[0,y[0]], LP[1,y[1]], LP[2,y[2]], ...,
#    # LP[n-1,y[n-1]]] and T.mean(LP[T.arange(y.shape[0]),y]) is
#    # the mean (across minibatch examples) of the elements in v,
#    # i.e., the mean log-likelihood across the minibatch.
#    # print "at least, y must be provided in a flattened view (a list of class values)!"
#     #[T.arange(y.shape[0]), y] == [ (0,1,2,3,4,...),(1,0,1,1,0,....) ]
#    #                                               select in each row the corresponding class entry
#    cls      = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x')  # (1,4,1, 1 )
#    y        = y.dimshuffle(0, 'x', 1, 2)                                                # (1,1,10,10)
#    select   = T.eq(cls, y).nonzero()
#    nll_inst = -T.log(self.class_probabilities)[select]
#    nll      = T.mean(nll_inst)
#
#    nll_inst = nll
#    return nll, nll_inst
#
#  def nll_classwise_masking(self, y, mask_class_labeled, mask_class_not_present):
#    """
#    Special Training objective function for lazy labelling
#
#    :type y: theano.tensor.TensorType
#    :param y: true classes (as integer values from 0 to (N_classes-1) ), shape = (batchsize, x, y)
#
#
#    :mask_class_labeled:
#      shape = (batchsize, num_classes).
#      A binary mask indicating whether a class is labeled in ``y``. If an class ``k``
#      is (in general) visible in the image data **and** ``mask_class_labeled[k]==1``, then
#      the labels  **must** obey``y=k`` in all pixels where it is visible.
#      If a class ``k`` is visible in the image, but was not labeled (to save time), set the
#      ``mask_class_labeled[k]=0``: if the CNN predicts classes which are
#      masked to zero it will not incur any loss. Alternative: set ``y=-1`` to the value -1 in areas which shall
#      be exempt from Training.
#      Limit case: ``mask_class_labeled[:]==1`` will result in the ordinary NLL.
#
#    :mask_class_not_present:
#      shape = (batchsize, num_classes).
#      A binary mask indicating whether a class is visible in the image data.
#      ``mask_class_not_present[k]==1`` means that the image does **not** contain examples of this class (irrespective of
#      the contents of <y> or <mask_class_labeled>).\
#      
#      Exemplary use case: set all <y>==-1 or <mask_class_labeled>==0 and indicate which classes should NOT
#      be predicted by the classifier by setting the corresponding entries of <mask_class_not_present> to 1.
#      This will push the predicted probabilities for those classes to zero.
#      Limit case: ``mask_class_labeled[:]==0`` will result in the ordinary NLL.
#
#    values of -1 in <y> count as "not labeled / ignore predictions"; \
#    but this has NO PRIORITY over <mask_class_not_present>.
#    """
#    y         = y.dimshuffle(0, 'x', 1, 2)  #(batchsize, 1, x, y)
#    cls       = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x') # available classes
#
#    mask_class_labeled     = mask_class_labeled.dimshuffle(0, 1, 'x', 'x')     #(batchsize, num_classes,1 ,1)
#    mask_class_not_present = mask_class_not_present.dimshuffle(0, 1, 'x', 'x') #(batchsize, num_classes,1 ,1)
#    # Mask of unlabeled positions.
#    # Apply to loss in order to ignore examples with label -1
#    label_avail_mask = T.neq(y,-1)
#
#    pred = self.class_probabilities # (batchsize, num_classes, x, y)
#    # Mask of unlabelled positions but with value 0 where stuff is not labelled
#    mod_y = T.where(y<0,0,y) # turns -1 in y to 0!
#    # Binary mask with value of 1 at the position of the labeled class (0 for other classes)
#    class_selection   = T.eq(cls, mod_y)
#    # dirty hack: compute "standard" nll when most predictive weight is put on classes which are in fact labeled
#    # as indicated in <mask_class_labeled>; threshold is 50%
#    votes_for_labeled = T.where( T.sum(pred*mask_class_labeled, axis=1)>=0.5, 1, 0 ).dimshuffle(0,'x',1,2)
#
#    # could also add '* mask_class_labeled' inside, but this should not change anything , provided there is
#    # no logical conflict between y and mask_class_labeled !
#    # standard loss part -> increase p(correct_prediction); thus disabled if the "correct" class is not known
#    nll_inst_up = -((T.log(pred) * votes_for_labeled * label_avail_mask * class_selection).sum(axis=1))
#    # multiply <pred> with <label_avail_mask> if <y>==-1 should override 'unlabeled' areas (currently not the case).
#    nll_inst_dn = -(T.log(1 - (pred * mask_class_not_present).sum(axis=1)))
#
#    nll_inst = nll_inst_up + nll_inst_dn
#    nll      = T.mean(nll_inst)
#    return nll, nll_inst

    def NLL(self,
            y,
            class_weights=None,
            mask_class_labeled=None,
            mask_class_not_present=None,
            label_prop_thresh=None):
        """
        Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
        of this model under a given target distribution.

        y: theano.tensor.TensorType
          corresponds to a vector that gives for each example the correct label. Labels < 0 are ignored (e.g. can
          be used for label propagation)

        class_weights: theano.tensor.TensorType
          weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
          ``w > 1.0`` (more important class)

        The following refers to lazy labels, the masks are always on a per patch basis, depending on the
        origin cube of the patch. The masks are properties of the individual image cubes and must be loaded
        into CNNData.

        mask_class_labeled: theano.tensor.TensorType
          shape = (batchsize, num_classes).
          Binary masks indicating whether a class is properly labeled in ``y``. If a class ``k``
          is (in general) present in the image patches **and** ``mask_class_labeled[k]==1``, then
          the labels  **must** obey ``y==k`` for all pixels where the class is present.
          If a class ``k`` is present in the image, but was not labeled (-> cheaper labels), set
          ``mask_class_labeled[k]=0``. Then all pixels for which the ``y==k`` will be ignored.
          Alternative: set ``y=-1`` to ignore those pixels.
          Limit case: ``mask_class_labeled[:]==1`` will result in the ordinary NLL.

        mask_class_not_present: theano.tensor.TensorType
          shape = (batchsize, num_classes).
          Binary mask indicating whether a class is present in the image patches.
          ``mask_class_not_present[k]==1`` means that the image does **not** contain examples of class ``k``.
          Then for all pixels in the patch, class ``k`` predictive probabilities are trained towards ``0``.
          Limit case: ``mask_class_not_present[:]==0`` will result in the ordinary NLL.

        label_prop_thresh: float (0.5,1)
          This threshold allows unsupervised label propagation (only for examples with negative/ignore labels).
          If the predictive probability of the most likely class exceeds the threshold, this class is assumed to
          be the correct label and the training is pushed in this direction.
          Should only be used with pre-trained networks, and values <= 0.5 are disabled.

        Examples:

        - A cube contains no class ``k``. Instead of labelling the remaining classes they can be
          marked as unlabelled by the first mask (``mask_class_labeled[:]==0``, whether ``mask_class_labeled[k]``
          is ``0`` or ``1`` is actually indifferent because the labels should not be ``y==k`` anyway in this case).
          Additionally ``mask_class_not_present[k]==1`` (otherwise ``0``) to suppress predictions of ``k`` in
          in this patch. The actual value of the labels is indifferent, it can either be ``-1`` or it could be the
          background class, if the background is marked as unlabelled (i.e. then those labels are ignored).

        - Only part of the cube is densely labelled. Set ``mask_class_labeled[:]=1`` for all classes, but set the
          label values in the un-labelled part to ``-1`` to ignore this part.

        - Only a particular class ``k`` is labelled in the cube. Either set all other label pixels to ``-1`` or the
          corresponding flags in ``mask_class_labeled`` for the unlabelled classes.

        ..  Note::
          Using ``-1`` labels or telling that a class is not labelled, is somewhat redundant and just
          supported for convenience.
        """

        # NOTE: This whole function has a ugly problem with NaN. They arise for pred values close to 0 or 1
        # (i.e. for NNs that make very confident and usually also correct predictions) because initially the log of
        # all the whole pred tensor is taken. Later we want to use only some indices of the tensor (mask) but
        # that is not so easy done. There are two ways:
        # 1. advanced indexing: T.log(pred)[mask.nonzero()] --> fails if mask is all zero, cannot be fixed
        # 2. mutiplying with 0-1-mask: T.log(pred) * mask.nonzero --> but NaN * 0 = NaN, but we require 0!
        # For the second option, in principle, the NaNs could be replaced by 0 using T.switch, but then the gradient
        # fails because the replaced value is disconnected from the parameters and gives NaN (mathematically
        # the gradient should correctly be 0 then; there is a Theano ticket open to request a fix).
        # So finally the best practice is to add a stabilisation to the log: T.log(pred) --> T.log(pred+eps)
        # This looks ugly, but does the task and the introduced error is completely negligible.

        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl, x, y)
        y = y.dimshuffle(0, 'x', 1, 2)  # the labels (bs, 1,  x, y)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x')  # available classes
        label_selection = T.eq(cls, y)  # selects correct labels

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0, 'x', 'x')

            # Up vote block
        if mask_class_labeled is not None:  # Standard
            mask_class_labeled = mask_class_labeled.dimshuffle(0, 1, 'x', 'x')
            label_selection = (label_selection * mask_class_labeled)

        nll_inst_up = -T.log(pred + eps) * label_selection * class_weights
        N_up = T.sum(label_selection)  # number of labelled examples

        if label_prop_thresh is not None:  # Label propagation block
            above_thresh = pred > label_prop_thresh  # this is one for the class with highes prob
            prop_mask = above_thresh * (1 - mask_class_labeled)  # don't do where training labels are available
            nll_inst_up_prop = -T.log(pred + eps) * prop_mask * class_weights
            N_up_prop = prop_mask.sum()

            nll_inst_up += nll_inst_up_prop
            N_up += N_up_prop

        if mask_class_not_present is not None:  # Down vote block
            mask_class_not_present = mask_class_not_present.dimshuffle(0, 1, 'x', 'x')
            nll_inst_dn = -T.log(1.0 - (pred * mask_class_not_present * class_weights) + eps)
            N_dn = mask_class_not_present.sum()  # number of not present classes examples
        else:
            nll_inst_dn = 0.0
            N_dn = 0.0

        nll_inst = (nll_inst_up + nll_inst_dn).sum(axis=1)
        N_total = (N_up + N_dn)
        # patch N_total to be not 0, when this is the case the sum is 0 anyway:
        N_total = T.switch(T.eq(N_total, 0), 1, N_total)
        nll = (nll_inst_up + nll_inst_dn).sum() / N_total

        return nll, nll_inst

    def NLL_weak(self,
                 y,
                 class_weights=None,
                 mask_class_labeled=None,
                 mask_class_not_present=None,
                 label_prop_thresh=None):
        """
        NLL that mixes the current cnn output and the hard labels as target
        """
        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl, x, y)
        y = y.dimshuffle(0, 'x', 1, 2)  # the labels (bs, 1,  x, y)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x')  # available classes
        hard_labels = T.eq(cls, y)  # selects correct labels

        soft_labels = 0.5 * hard_labels + 0.5 * pred

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0, 'x', 'x')

        nll_inst_up = -(soft_labels * T.log(pred + eps)) * class_weights
        #nll_inst_up = - 0.5 * (hard_labels * T.log(pred + eps)) - 0.5 * (pred * T.log(pred + eps))
        #nll_inst_up = -T.log(pred + eps)  * class_weights
        N_up = T.sum(hard_labels)  # number of labelled examples

        nll_inst_dn = 0.0
        N_dn = 0.0

        N_total = (N_up + N_dn)
        # patch N_total to be not 0, when this is the case the sum is 0 anyway:
        N_total = T.switch(T.eq(N_total, 0), 1, N_total)
        nll = (nll_inst_up + nll_inst_dn).sum() / N_total

        return nll, nll

    def squared_distance(self, y):
        """
        Returns squared distance between prediction and ``y``

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                      correct label
        """
        return T.mean((self.output - y)**2)

    def errors(self, y):
        """
        Returns classification accuracy

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label
        """
        # check if y has same dimension of y_pred
        if y.ndim != self.class_prediction.ndim:
            raise TypeError('y should have the same shape as self.class_prediction',
                            ('y', y.type, 'class_prediction', self.class_prediction.type))
        # check if y is of the correct datatype
        if y.dtype.startswith('int'):
            return T.mean(T.neq(self.class_prediction, y)[T.ge(y, 0).nonzero()])
        else:
            print "something went wrong"
            raise NotImplementedError()

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import time
import numpy as np
import theano
import theano.tensor as T
import matplotlib.pyplot as plt
from scipy.optimize import fmin_l_bfgs_b
from collections import OrderedDict


class Optimizer(object):
    """
    Optimizer Base Object, initialises generic optimizer variables

    Parameters
    ----------

    model_obj: cnn-object
      Encapsulation of theano model (instead of giving X,Y etc. manually), all other arguments are
      retrieved from this object if they are ``None``. If an argument is not ``None`` it will override the
      value from the model
    X:         symbolic input variable
      Data
    Y:         symbolic output variable
      Target
    Y_aux:     symbolic output variable
      Auxiliary masks/weights/etc. for the loss, type: list!
    top_loss:  symbolic loss function:
      Requires (X, Y (,*Y_aux)) for compilation
    params:    list of shared variables
      List of parameter arrays against which the loss is optimised

    Returns
    -------
    Callable optimizer object: loss = Optimizer(X, Y (,*Y_aux)) performs one iteration

    """

    def __init__(self,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None):
        self.t_init = time.time()
        self.optimizer_params = None
        self.step = None
        if model_obj is None:
            assert X is not None and Y is not None and top_loss is not None and params is not None, "Missing arguments!"
            self.gradients = T.grad(top_loss,
                                    params,
                                    disconnected_inputs="warn")

        else:
            if X is None:
                self.X = model_obj._x
            else:
                self.X = X
            if Y is None:
                self.Y = model_obj._y
            else:
                self.Y = Y
            if Y_aux == []:
                self.Y_aux = model_obj._y_aux
            else:
                self.Y_aux = Y_aux
            if params is None:
                self.params = model_obj.params
            else:
                self.params = params

            if top_loss is None:  # no loss means, loss&grad must already be defined in the model
                self.gradients = model_obj._gradients
                self.top_loss = model_obj._loss

            else:
                self.gradients = T.grad(top_loss, params, disconnected_inputs="warn")

        assert len(self.params) > 0, "no params, call compileOutputFunctions() before calling compileOptimizer()!"

        if hasattr(model_obj, '_last_grads') and model_obj._last_grads != [] and model_obj._last_grads is not None:
            self.last_grads = model_obj._last_grads
        else:
            self.last_grads = []
            for i, p in enumerate(self.params):
                if p in self.params[:i]:
                    print "Detected shared param: param[%i]" % i
                else:
                    self.last_grads.append(theano.shared(np.zeros(p.get_value().shape, dtype=theano.config.floatX),
                                                         name=p.name + str('_LG'), borrow=False))

        if hasattr(model_obj, 'global_weightdecay'):
            self.weightdecay = model_obj.global_weightdecay
        else:
            self.weightdecay = 0

        if hasattr(model_obj, '_loss_instance'):
            self.loss_instance = model_obj._loss_instance
        else:
            self.loss_instance = None

        if hasattr(model_obj, '_get_loss'):
            self.get_loss = model_obj._get_loss
        else:
            self._get_loss = theano.function([self.X, self.Y] + self.Y_aux, [self.top_loss, self.loss_instance])

        if model_obj is not None:
            model_obj._last_grads = self.last_grads  # share last grads in model
            model_obj.get_loss = self.get_loss
            self.model_obj = model_obj

    def updateOptimizerParams(self, optimizer_params):
        """
        Update the hyper-parameter dictionary
        """
        self.optimizer_params.update(optimizer_params)

    def get_loss(self, *args):
        """
        [data, labels(, *aux)] --> [loss, loss_instance]
        loss_instance is the loss per instance (e.g. batch-item or pixel)
        """
        loss, nloss_instance = self._get_loss(*args)
        return np.float32(loss), nloss_instance

    def __call__(self, *args):
        """
        Perform an update step

        [data, labels(, *aux)] --> [loss, loss_instance]
        loss_instance is the loss per instance (e.g. batch-item or pixel)
        """
        ret = list(self.step(*args))
        ret[0] = np.float32(ret[0])  # the scalar nll
        return ret

    def compileGradients(self):
        """
        Compile and return a function that returns list of gradients
        """
        print "Compiling getGradients"
        getGradients = theano.function(
            [self.X, self.Y] + self.Y_aux,
            self.gradients,
            on_unused_input='warn')
        return getGradients

##############################################################################################################
#######################################       Stochastic Gradient Descent     ################################
##############################################################################################################


class compileSGD(Optimizer):
    """
    Stochastic Gradient Descent
    """

    def __init__(self,
                 optimizer_params,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None):
        print "Compiling SGD"
        super(compileSGD, self).__init__(model_obj, X, Y, Y_aux, top_loss, params)

        self.LR = optimizer_params['LR']
        self.momentum = optimizer_params['momentum']

        grad_updates = []
        param_updates = []
        for param_i, grad_i, last_grad_i in zip(self.params, self.gradients, self.last_grads):
            new_grad_i = grad_i + last_grad_i * self.momentum
            grad_updates.append((last_grad_i, new_grad_i))  # use this if you want to use the gradient magnitude
            # For no weight decay weightdecay is just 0
            param_updates.append((param_i, param_i - (self.LR) * new_grad_i - self.LR * self.weightdecay * param_i))

        assert len(grad_updates) == len(param_updates), str(len(grad_updates)) + " != " + str(len(param_updates))
        # This updates last_grads with the current grad and returns the loss before any parameter change
        self.step = theano.function(
            [self.X, self.Y] + self.Y_aux,
            [self.top_loss, self.loss_instance],
            updates=param_updates + grad_updates,
            on_unused_input='warn')

        print " Compiling done  - in %.3f s!" % (time.time() - self.t_init)

##############################################################################################################
#######################################                Ada Grad               ################################
##############################################################################################################


class compileAdam(Optimizer):
    """
    Stochastic Gradient Descent
    """

    def __init__(self,
                 optimizer_params,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None):
        print "Compiling Adam"
        super(compileAdam, self).__init__(model_obj, X, Y, Y_aux, top_loss, params)

        self.LR = optimizer_params['LR']
        beta1 = optimizer_params['beta1']
        beta2 = optimizer_params['beta2']
        epsilon = optimizer_params['epsilon']

        updates = OrderedDict()
        t_prev = theano.shared(np.float32(0.0))
        t = t_prev + 1.0
        a_t = self.LR * T.sqrt(1 - beta2**t) / (1 - beta1**t)

        for param, g_t in zip(self.params, self.gradients):
            value = param.get_value(borrow=True)
            m_prev = theano.shared(np.zeros(value.shape, dtype=value.dtype), broadcastable=param.broadcastable)
            v_prev = theano.shared(np.zeros(value.shape, dtype=value.dtype), broadcastable=param.broadcastable)

            m_t = beta1 * m_prev + (1 - beta1) * g_t
            v_t = beta2 * v_prev + (1 - beta2) * g_t**2
            step = a_t * m_t / (T.sqrt(v_t) + epsilon)

            updates[m_prev] = m_t
            updates[v_prev] = v_t
            updates[param] = param - step

        updates[t_prev] = t
        self.step = theano.function(
            [self.X, self.Y] + self.Y_aux,
            [self.top_loss, self.loss_instance],
            updates=updates,
            on_unused_input='warn')

        print " Compiling done  - in %.3f s!" % (time.time() - self.t_init)

##############################################################################################################
#######################################       Resilient backPROPagation      #################################
##############################################################################################################


class compileRPROP(Optimizer):
    """ Resilient backPROPagation """

    def __init__(self,
                 optimizer_params,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None):
        print "Compiling RPROP..."
        super(compileRPROP, self).__init__(model_obj, X, Y, Y_aux, top_loss,
                                           params)

        self.LRs = []
        RPROP_updates = []

        # Initialise shared variables for the Training algos
        for i, para in enumerate(self.params):
            if para in self.params[:i]:
                print "Detected RNN or shared param @index =", i
            else:
                self.LRs.append(theano.shared(
                    np.float32(optimizer_params['initial_update_size']) *
                    np.ones(para.get_value().shape, dtype=theano.config.floatX),
                    name=para.name + str('_RPROP'),
                    borrow=0))

        print "RPROP: missing backtracking handling "  ###TODO ???
        for param_i, grad_i, last_grad_i, pLR_i in zip(
                self.params, self.gradients, self.last_grads, self.LRs):
            # Commented code on next 4 lines is theano-incapable and just illustration!!!
            # if   ((last_grad_i*grad_i) < -1e-9): # sign-change & significant magnitude of last two gradients
            #   pLR_i_new = pLR_i * (1 - np.float32(RPROP_penalty)) # decrease this LR
            # elif ((last_grad_i*grad_i) > 1e-11): # no sign-change & and last two gradients were sufficiently big
            #   pLR_i_new = pLR_i * (1 + np.float32(RPORP_gain))    # increase this LR

            # capping RPROP-LR inside [1e-7,2e-3]
            RPROP_updates.append((pLR_i, T.minimum(T.maximum(
                    pLR_i * (1 - np.float32(optimizer_params['penalty']) * ((last_grad_i * grad_i) < -1e-9)
                             + np.float32(optimizer_params['gain']) * ((last_grad_i * grad_i) > 1e-11)),
                    1e-7 * T.ones_like(pLR_i)), 2e-3 * T.ones_like(pLR_i))))
            RPROP_updates.append((param_i, param_i - pLR_i * grad_i / (T.abs_(
                grad_i) + 1e-6) - (self.weightdecay * param_i)))
            RPROP_updates.append((last_grad_i, grad_i))

        self.step = theano.function([self.X, self.Y] + self.Y_aux,
                                    [self.top_loss, self.loss_instance],
                                    updates=RPROP_updates,
                                    on_unused_input='warn')
        print " Compiling done  - in %.3f s!" % (time.time() - self.t_init)

##############################################################################################################
###########################################       Conjugate Gradient      ####################################
##############################################################################################################


class compileCG(Optimizer):
    """ Conjugate Gradient """

    def __init__(self,
                 optimizer_params,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None):
        print "Compiling CG"
        super(compileCG, self).__init__(model_obj, X, Y, Y_aux, top_loss,
                                        params)

        self.optimizer_params = optimizer_params
        self.direc = []
        CG_updates_direc = []
        CG_updates_grads = []
        CG_updates = []  # params
        self.model_obj = model_obj

        # Initialise shared variables for the Training algos
        for para in self.params:
            para_shape = para.get_value().shape
            self.direc.append(theano.shared(
                np.zeros(para_shape, dtype=theano.config.floatX),
                name=para.name + '_CG_direc',
                borrow=False))

        ### Kickstart of CG, initialise first direction to current gradient ###
        for grad_i, last_grad_i, direc_i in zip(self.gradients, self.last_grads, self.direc):
            CG_updates_grads.append((last_grad_i, grad_i))
            CG_updates_direc.append((direc_i, -grad_i))

        # update direc & last-grad
        updates = CG_updates_grads + CG_updates_direc
        self.CG_kickstart = theano.function(
            [self.X, self.Y] + self.Y_aux,
            None,
            updates=updates,
            on_unused_input='ignore')

        ### Regular CG-step ###
        CG_updates_direc = []  # clear update-list
        CG_updates_grads = []  # clear update-list

        # Compute Polak-Ribiere coefficient b for updating search direction
        denom = theano.shared(np.float32(0))
        num = theano.shared(np.float32(0))

        for grad_i, last_grad_i in zip(self.gradients, self.last_grads):
            num = num + T.sum(-grad_i * (-grad_i + last_grad_i))
            denom = denom + T.sum(last_grad_i * last_grad_i)
        coeff = num / denom
        coeff = T.max(T.stack([coeff, theano.shared(np.float32(0))]))  # select

        # Search-direction and last-grad update
        for grad_i, last_grad_i, direc_i in zip(self.gradients, self.last_grads, self.direc):
            CG_updates_grads.append((last_grad_i, grad_i))
            CG_updates_direc.append((direc_i, -grad_i + direc_i * coeff))

        updates = CG_updates_grads + CG_updates_direc
        #    if self.Y_aux is None:
        #      self.CG_step = theano.function([self.X, self.Y],
        #                                     coeff, updates=updates, on_unused_input='ignore')
        #    else:
        self.CG_step = theano.function([self.X, self.Y] + self.Y_aux, coeff, updates=updates, on_unused_input='ignore')

        # Weights update (Line search), no input needed, as only params are changed
        delta = T.fscalar('delta')  # used to parametrise the ray along we search (=0 at current params)
        self.t = theano.shared(np.float32(0))  # Internal update step indicator
        for param_i, search_direc_i in zip(self.params, self.direc):
            CG_updates.append((param_i, param_i + search_direc_i * delta))

        CG_updates.append((self.t, self.t + delta))
        self.CG_update_params = theano.function([delta], self.t + delta, updates=CG_updates)

        # Linear-Approximation (from shared last_(grad|direc))
        linear_approx = theano.shared(np.float32(0))
        for grad_i, last_direc_i in zip(self.last_grads, self.direc):
            linear_approx = linear_approx + T.sum(grad_i * last_direc_i)

        self.CG_linear_approx = theano.function([], linear_approx, updates=None)

        print " Compiling done  - in %.3f s!" % (time.time() - self.t_init)

    def __call__(self, *args):  # i.e. trainingStepCG
        """
        Perform an update step

        [data, labels(, *aux)] --> [loss, loss_instance]
        loss_instance is the loss per instance (e.g. batch-item or pixel)
        """
        self.CG_kickstart(*args)
        timeline = []
        loss, loss_instance, t, count = self.lineSearch(*args)
        timeline.append([loss, t, 0, count])
        self.t.set_value(np.float32(0))  # DBG: reset internal update-magnitude
        n_steps = self.optimizer_params['n_steps']
        for i in xrange(n_steps - 1):
            if self.optimizer_params['only_descent']:
                self.CG_kickstart(*args)
                coeff = 0
            else:  # use actual CG
                coeff = self.CG_step(*args)

            loss, loss_instance, t, count = self.lineSearch(*args)
            self.t.set_value(np.float32(0))
            timeline.append([loss, t, coeff, count])

        if hasattr(self.model_obj, 'CG_timeline'):
            self.model_obj.CG_timeline.extend(timeline)
        return loss, loss_instance

    def lineSearch(self, *args):
        """ Needed for CG """
        loss_0, _ = self.get_loss(*args)
        loss_0 = np.float32(loss_0)
        linear_approx = self.CG_linear_approx()
        counter = 0
        if linear_approx > 0:  # if algorithm gets stuck, reset
            loss, loss_instance = self.get_loss(*args)
            return loss_0, loss_instance, 0, counter
        max_step = self.optimizer_params['max_step']
        min_step = self.optimizer_params['min_step']
        beta = self.optimizer_params['beta']

        max_count = int(np.log(min_step / (max_step)) / np.log(beta))  # limit iterations to lower bound
        points = []  ### For Plotting
        t = max_step
        # The next search points DEcrement the search ray by the decaying negative factor delta
        # Thus parameters needn't to be reset before updating, we directly change
        # the parameters by the desired amount
        # The first search point: max_step along the current search direction
        self.CG_update_params(np.float32(max_step))
        loss, loss_instance = self.get_loss(*args)
        loss = np.float32(loss)
        counter += 1
        delta = max_step * (beta - 1.0)
        last_loss = 1000000
        for i in xrange(max_count):
            chord = loss_0 + t * linear_approx * self.optimizer_params['alpha']
            points.append([t, loss, chord])
            # The second condition is a deviation from regular BT-LineSearch
            if (loss <= chord) or (loss > last_loss):
                break
            self.CG_update_params(np.float32(delta))
            last_loss = loss
            loss, loss_instance = self.get_loss(*args)
            loss = np.float32(loss)
            delta = delta * beta
            t = t * beta
            counter += 1

        if self.optimizer_params['show']:
            points.append([0, loss_0, loss_0])
            points = np.array(points)
            plt.figure()
            plt.plot(points[:, 0], points[:, 1:])
            plt.scatter(points[:, 0], points[:, 1])
            plt.legend(('fu', 'chord'))
            plt.draw()
            plt.pause(0.0001)

        return loss, loss_instance, np.float(self.t.eval()), counter

##############################################################################################################
###########################################             L-BFGS            ####################################
##############################################################################################################


class compileLBFGS(Optimizer):
    """
    L-BFGS (fast, full-batch method)

    References (cite one):
      R. H. Byrd, P. Lu and J. Nocedal. A Limited Memory Algorithm for Bound Constrained Optimization,
      (1995), SIAM Journal on Scientific and Statistical Computing, 16, 5, pp. 1190-1208.

      C. Zhu, R. H. Byrd and J. Nocedal. L-BFGS-B: Algorithm 778: L-BFGS-B, FORTRAN routines for large scale
      bound constrained optimization (1997), ACM Transactions on Mathematical Software, 23, 4, pp. 550 - 560.

      J.L. Morales and J. Nocedal. L-BFGS-B: Remark on Algorithm 778: L-BFGS-B, FORTRAN routines for large
      scale bound constrained optimization (2011), ACM Transactions on Mathematical Software, 38, 1.

    """

    def __init__(self,
                 optimizer_params,
                 model_obj=None,
                 X=None,
                 Y=None,
                 Y_aux=[],
                 top_loss=None,
                 params=None,
                 debug=False):
        print "Compiling lBFGS"
        super(compileLBFGS, self).__init__(model_obj, X, Y, Y_aux, top_loss, params)

        self.optimizer_params = optimizer_params
        ret = [self.top_loss] + self.gradients
        self._loss_and_grad = theano.function([self.X, self.Y] + self.Y_aux, ret, on_unused_input='warn')

        self.params_values = [p.get_value() for p in self.params]
        self.shapes = [p.shape for p in self.params_values]  # list of param-shapes
        self.sizes = map(np.prod, self.shapes)  # list of param-sizes
        self.total_size = np.sum(self.sizes)  # length of vectorized parameters
        self.params_vec = np.zeros(self.total_size, "float32")
        self.gradients_vec = np.zeros(self.total_size, "float32")

        self.debug = debug
        print " Compiling done  - in %.3f s!" % (time.time() - self.t_init)

    def vec2list(self, vec, target_list):
        pos = 0
        for a, shape, size in zip(target_list, self.shapes, self.sizes):
            a[...] = vec[pos:pos + size].reshape(shape)
            pos += size

    def list2vect(self, src_list, target_vec):
        pos = 0
        for a, size in zip(src_list, self.sizes):
            target_vec[pos:pos + size] = np.asarray(a).flatten()
            pos += size

    def loss_and_grad(self, params_vect_new, *args):
        """ internal use, updates self.params"""
        self.vec2list(params_vect_new, self.params_values)  # Update param values in list and then on GPU
        for p, val in zip(self.params, self.params_values):
            p.set_value(val, borrow=False)

        ret = self._loss_and_grad(*args)
        loss, grads = np.float32(ret[0]), ret[1:]
        if len(grads) == 1:
            grads = grads[0]

        self.list2vect(grads, self.gradients_vec)
        grad_vec = np.asarray(self.gradients_vec, dtype="float64")

        if self.debug:
            print "loss_and_grad:: x_min,max =",np.min(args[0]),np.max(args[0]),\
            "y_min,max =",np.min(args[1]),np.max(args[1])
            print "loss_and_grad (av(g) =", np.mean(
                grad_vec), ", av(abs(g)) =", np.mean(np.abs(grad_vec)), ")"

        return loss, grad_vec

    def __call__(self, *args):
        """
        Perform an update step

        [data, labels(, *aux)] --> [loss, loss_instance]
        loss_instance is the loss per instance (e.g. batch-item or pixel)
        """
        if self.debug:
            print "\nlbfgs->optimize::"
            print "__optimize:: x_min,max =", np.min(args[0]), np.max(args[0]), "y_min,max =",\
                  np.min(args[1]), np.max(args[1])

        self.params_values = [p.get_value() for p in self.params]
        self.list2vect(self.params_values, self.params_vec)

        params, loss, info_dict = fmin_l_bfgs_b(func=self.loss_and_grad,
                                                x0=self.params_vec,
                                                args=args,
                                                **self.optimizer_params)
        aborted = info_dict["warnflag"] != 1
        if aborted:
            print info_dict
            print "L-BFGS aborted!"
        return loss



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

print "Load ELEKTRONN Core"

import time, sys
import cPickle
import numpy as np
import theano
import theano.tensor as T

from elektronn.utils import pprinttime

import optimizer as opt
from perceptronlayer import PerceptronLayer, RecurrentLayer
from convlayer2d import ConvLayer2d
from convlayer3d import ConvLayer3d, AffinityLayer3d, MalisLayer


def _printOps(n):
    """
    Return a humanized string representation of a large number.
    """
    abbrevs = ((1000000000000, 'Tera Ops'), (1000000000, 'Giga Ops'), (1000000, 'Mega Ops'), (1000, 'kilo Ops'))
    for factor, suffix in abbrevs:
        if n >= factor:
            break
    print 'Computational Cost: %.1f %s' % (float(n) / factor, suffix)


class MixedConvNN(object):
    """
    Parameters
    ----------

    input_size: tuple
    Data shapes, excluding batch and channel (used to infer the dimensionality)
    input_depth: int/None
    Is None by default this means non-image data (no conv layers allowed). Change to 1 for b/w, 3 for RGB and\
    4 for RGB-D images etc. For RNN this is the length of the time series.
    batch_size: int/None
    None for variable batch size
    enable_dropout:  Bool
    Turn on or off dropout
    recurrent:       Bool
    Support recurrent iterations along input depth/time
    dimension_calc: dimension calculator object

    Examples
    --------

    Note that image data must have at least 1 channel, e.g. a 2d image (1,x,y). 3d requires data in the
    format (z,ch,x,y). E.g. to create an isotropic 3d CNN with 5 channels (total input shape is (1,30,5,30,30)):

    >>> MixedConvNN((30,30,30), input_depth=5, batch_size=1)

    A non-convolutional MLP can be created as:

    >>> MixedConvNN((100,), input_depth=None, batch_size=2000)

    """

    def __init__(self,
                 input_size=None,
                 input_depth=None,
                 batch_size=None,
                 enable_dropout=False,
                 recurrent=False,
                 dimension_calc=None):
        assert input_size is not None
        self.layers = []  # [0] input layer ---> [-1] output layer
        self.poolings = []
        self.params = []

        self._output_layers = []  # Empty UNLESS you use add sth explicitly
        self._autoencoder_chains = []
        self._last_grads = []
        self.debug_functions = []
        self.debug_conv_output = []
        self.debug_gradients_function = []

        self.CG_timeline = []
        self.batch_size = batch_size
        self.n_lab = None
        self.input_shape = None
        self.patch_size = np.array(input_size
                                   )  # onlt the spatial part of input shape
        self.output_strides = None
        self.output_shape = None
        self.mfp_strides = None
        self.mfp_offsets = None
        self.dimension_calc = dimension_calc
        self.TotalForwardPassCost = 0  # number of multiplications done
        self.SGD_LR = theano.shared(np.float32(0.09))  # those 3 values are to be overwritten
        self.SGD_momentum = theano.shared(np.float32(0.9))
        self.global_weightdecay = theano.shared(np.float32(0))
        self._SGD_params = {'LR': self.SGD_LR, 'momentum': self.SGD_momentum}
        self._RPROP_params = {}
        self._CG_params = {}
        self._LBFGS_params = {}
        self._use_class_weights = False
        self._enable_dropout = enable_dropout
        self._recurrent = recurrent
        self._atleast_single_mfp = False
        self._y = None
        self._y_aux = []

        self.input_noise = None
        self.t_init = time.time()

        try:
            input_size = tuple(input_size, )
        except:
            input_size = (input_size, )
        input_dim = len(input_size)
        self.n_dim = input_dim
        assert input_dim in [
            1, 2, 3
        ], "MixedConvNN: input_dimension currently not supported"

        if input_dim > 1 and input_depth is None:
            input_depth = 1
            print "For image-like data no depth was specified, using depth=1"

        if recurrent:
            assert input_dim == 1
            self._x = T.ftensor3('x_rnn_input')
            if input_depth is not None:
                self.input_shape = (batch_size, input_depth, input_size[0])  # [batch, time, feat]
            else:
                self.input_shape = (batch_size, input_size[0])  # the input is repeated (see "iterations" in rnn layer)
        else:
            x_dim = input_dim + 1  # +1 because of leading batch dimension
            if input_depth is not None:  # For images there is always an additional channel dimension (even if it is 1)
                x_dim += 1
                self.input_shape = (batch_size, input_depth) + input_size
            else:  # For non image input / prohibits ConvLayers
                self.input_shape = (batch_size, ) + input_size

            # construct tensor of matching dimensionality
            self._x = T.TensorType('float32', (False,) * x_dim, name='x_cnn_input')()
            if input_dim == 3:  # strange order for theano 3dconv
                self.input_shape = (batch_size, input_size[0], input_depth, input_size[1], input_size[2])

        print '-' * 60
        print "Input shape   = ", self.input_shape, "; This is a", input_dim, "dimensional NN"
        if batch_size is not None:
            self._layer0_input = self._x.reshape(self.input_shape)
        else:
            self._layer0_input = self._x

        print '---'

    ############################################################################################################

    def addPerceptronLayer(self,
                           n_outputs=10,
                           activation_func='tanh',
                           enable_input_noise=False,
                           add_in_output_layers=False,
                           force_no_dropout=False,
                           W=None,
                           b=None):
        """
        Adds a Perceptron layer to the CNN.

        Normally the each layer creates its own set of randomly initialised neuron weights. To reuse the weights
        of another layer (weight sharing) use the arguments ``W`` and ``b`` an pass ``T.TensorVariable``.
        If ``W`` and ``b`` are numpy arrays own weights are initialised with these values.

        Parameters
        ----------

        n_outputs: int
          The size of this layer
        activation_func: string
          {tanh, relu, sigmoid, abs, linear, maxout <i>}
          Activation function
        enable_input_noise: Bool
          If True set 20% of input to 0 randomly (similar to dropout)
        force_no_dropout:   Bool
          Set True for last/output layer
        """
        layer_input_shape = self.input_shape if (
            self.layers == []) else self.layers[-1].output_shape
        layer_input = self._layer0_input if (
            self.layers == []) else self.layers[-1].output

        if len(layer_input_shape) > 2:  # input_dimension >= 2
            layer_input = layer_input.flatten(2)
            nin = (layer_input_shape[0], np.product(layer_input_shape[1:]))
        elif len(layer_input_shape) == 2:  # input_dimension = 1
            nin = layer_input_shape
        else:
            raise ValueError('Used invalid input dimension for Perceptron layer')

        input_noise = theano.shared(np.float32(0.2)) if enable_input_noise else None
        self.input_noise = input_noise if enable_input_noise else self.input_noise
        self._y = T.wvector('y_cnn_labels')
        layer = PerceptronLayer(
            input=layer_input,
            n_in=nin[1],
            n_out=n_outputs,
            batch_size=nin[0],
            enable_dropout=(self._enable_dropout and force_no_dropout == False),
            activation_func=activation_func,
            input_noise=input_noise,
            input_layer=self.layers[-1] if len(self.layers) > 0 else None,
            W=W,
            b=b)

        if add_in_output_layers:
            self._output_layers.append(layer)
        else:
            self.layers.append(layer)

        if self.batch_size is not None:
            num_multiplications = np.product(n_outputs) * np.product(layer_input_shape)
        else:
            num_multiplications = np.product(n_outputs) * np.product(layer_input_shape[1:])
        _printOps(num_multiplications)
        print '---'
        self.TotalForwardPassCost += num_multiplications

    ############################################################################################################

    def addConvLayer(self,
                     nof_filters=None,
                     filter_size=None,
                     pool_shape=2,
                     activation_func='tanh',
                     add_in_output_layers=False,
                     force_no_dropout=False,
                     use_fragment_pooling=False,
                     reshape=False,
                     is_last_layer=False,
                     layer_input_shape=None,
                     layer_input=None,
                     W=None,
                     b=None,
                     pooling_mode='max',
                     affinity=False):
        """
        Adds a convolutional layer to the CNN. The dimensionality is *automatically* inferred.

        Normally the inputs are automatically connected the the outputs of the last added layer. To connect to a
        different layer use ``layer_input_shape`` and ``layer_input`` arguments.

        Normally the each layer creates its own set of randomly initialised neuron weights. To reuse the weights
        of another layer (weight sharing) use the arguments ``W`` and ``b`` an pass ``T.TensorVariable``.
        If ``W`` and ``b`` are numpy arrays own weights are initialised with these values.

        Parameters
        ----------
        nof_filters:          int
          Number of feature maps
        filter_size:          int/tuple
          Size/shape of convolutional filters, xy/zxy, (scalars are automatically extended to the 2d or 3d)
        pool_shape:   int/tuple
          Size/shape of pool, xy/zxy, (scalars are automatically extended to the 2d or 3d)
        activation_func:      string
          {tanh, relu, sigmoid, abs, linear, maxout <i>}
          Activation function
        force_no_dropout:     Bool
          Set True for last/output layer
        use_fragment_pooling: Bool
         Set to True for predicting dense images efficiently. Requires batch_size==1.
        reshape:              Bool
          Set to True to get 2d/3d output instead of flattened class_probabilities in the last layer
        is_last_layer:        Bool
          Shorthand for reshape=True, force_no_dropout=True and reconstruction of pooling fragments (if mfp was active)
        layer_input_shape: tuple of int
          Only needed if layer_input is not not None
        layer_input: T.TensorVariable
          Symbolic input if you do *not* want to use the previous layer of the cnn. This requires specification of
          the shape of that input with ``layer_input_shape``.
        W: np.ndarray
          weight matrix. If array, the values are used to initialise a shared variable for this layer.
                               If TensorVariable, than this variable is directly used (weight sharing with the
                               layer from which this variable comes from)

        b: np.ndarray or T.TensorVariable
          bias vector. If array, the values are used to initialise a shared variable for this layer.
                               If TensorVariable, than this variable is directly used (weight sharing with the
                               layer from which this variable comes from)
        pooling_mode: str
          'max' or 'maxabs' where the first is normal maxpooling and the second also retains sign of large negative values
        """

        n_dim = self.n_dim
        assert n_dim in [2, 3], "only 2d and 3d convolution supported!"
        if not hasattr(filter_size, '__len__'):
            filter_size = (filter_size, ) * n_dim
        elif len(filter_size) == 1:
            filter_size = filter_size * n_dim
        elif len(filter_size) != n_dim:
            raise ValueError(
                'Filter size must be either scalar or have same length as n_dim')

        if not hasattr(pool_shape, '__len__'):
            pool_shape = (pool_shape, ) * n_dim
        elif len(pool_shape) == 1:
            pool_shape = pool_shape * n_dim

        self.poolings.append(pool_shape)

        if (layer_input_shape is None) and (layer_input is None):
            layer_input_shape = self.input_shape if len(self.layers) == 0 else self.layers[-1].output_shape
            layer_input = self._layer0_input if len(self.layers) == 0 else self.layers[-1].output
        else:
            assert (layer_input_shape is not None) and (layer_input is not None),\
                "Provide either both input and shape or neither"
        assert len(layer_input_shape) in [3,4,5],\
            "Please implement the stacking of a convLayer on top of PerceptronLayer (if this is your goal)"

        if is_last_layer:
            print "Last Layer, by default: no dropout and reshaped outputs"
            force_no_dropout = True
            reshape = True
            if self._atleast_single_mfp:
                use_fragment_pooling = True

        if use_fragment_pooling:
            if self.batch_size != 1:
                print("MFP is activated and batch_size is not 1")
                #raise ValueError("MFP is activated and batch_size is not 1")
                # self.batch_size = 1 doesn't help

                # if there is mfp in at least 1 layer the output must be reshaped
            self._atleast_single_mfp = use_fragment_pooling or self._atleast_single_mfp

        dropout = (self._enable_dropout and force_no_dropout == False)

        if n_dim == 2:
            filter_shape = (nof_filters, layer_input_shape[1], filter_size[0], filter_size[1])
            CL = ConvLayer2d
            if reshape:
                self._y = T.TensorType('int16', [False, False, False], name='y_cnn_labels')()

        if n_dim == 3:
            filter_shape = (nof_filters, filter_size[0], layer_input_shape[2], filter_size[1], filter_size[2])
            CL = ConvLayer3d

            if affinity:
                print "WARNING: hack for adding affinity layer / MALIS active"
                if affinity == 'malis':
                    CL = MalisLayer
                else:
                    CL = AffinityLayer3d

            if reshape:
                self._y = T.TensorType('int16', [False, False, False, False], name='y_cnn_labels')()

        layer = CL(
            layer_input,
            layer_input_shape,
            filter_shape,
            pool_shape,
            activation_func,
            dropout,
            use_fragment_pooling,
            reshape,
            self.mfp_offsets,
            self.mfp_strides,
            input_layer=self.layers[-1] if len(self.layers) > 0 else None,
            W=W,
            b=b,
            pooling_mode=pooling_mode)

        self.mfp_offsets = layer.mfp_offsets
        self.mfp_strides = layer.mfp_strides

        if add_in_output_layers:
            self._output_layers.append(layer)
        else:
            self.layers.append(layer)

        # Calculate computational cost
        if n_dim == 2:
            n_pos = ((layer_input_shape[2]+1-filter_size[0]) *\
                     (layer_input_shape[3]+1-filter_size[1]))
        if n_dim == 3:
            n_pos = ((layer_input_shape[1]+1-filter_size[0]) *\
                     (layer_input_shape[3]+1-filter_size[1]) *\
                     (layer_input_shape[4]+1-filter_size[2]))
        if self.batch_size is not None:
            num_multiplications = np.product(filter_size) * n_pos * nof_filters *\
                                  layer_input_shape[1 if n_dim==2 else 2] * layer_input_shape[0]
        else:
            num_multiplications = np.product(filter_size) * n_pos * nof_filters *\
                                  layer_input_shape[1 if n_dim==2 else 2] # Cost for 1 patch

        _printOps(num_multiplications)
        print "Param count:", layer.params[0].get_value().size, '+', layer.params[1].get_value().size, '=',\
              layer.params[0].get_value().size + layer.params[1].get_value().size
        print '---'
        self.TotalForwardPassCost += num_multiplications

    ############################################################################################################

    def addRecurrentLayer(self,
                          n_hid=None,
                          activation_func='tanh',
                          iterations=None):
        """
        Adds a recurrent layer (only possible for non-image input of format (batch, time, features))

        Parameters
        ----------

        n_hid: int
           Number of hidden units
        activation_func: string
          {tanh, relu, sigmoid, abs, linear}
        iterations: int
          If layer input is not time-like (iterable on axis 1) it can be broadcasted and
          iterated over for a fixed number of iterations
        """
        layer_input_shape = self.input_shape if (self.layers == []) else self.layers[-1].output_shape
        layer_input = self._layer0_input if (self.layers == []) else self.layers[-1].output
        # Padding of constant input
        if len(layer_input_shape) == 2:
            print "Recurrence with broadcasted input"
            assert isinstance(iterations, int)
            bs = layer_input_shape[0] if (layer_input_shape[0] is not None) else 1
            broadcaster = (bs, iterations, layer_input_shape[1])
            layer_input = layer_input.dimshuffle(0, 'x', 1) * T.ones(broadcaster, dtype=theano.config.floatX)
            layer_input_shape = (layer_input_shape[0], iterations, layer_input_shape[1])
        elif len(layer_input_shape) != 3:
            raise ValueError('Used invalid input dimension for Recurrent layer')

        nin = layer_input_shape  # [batch, time, features]

        self._y = T.wvector('y_cnn_labels')
        layer = RecurrentLayer(input=layer_input,
                               n_in=layer_input_shape[2],
                               n_hid=n_hid,
                               batch_size=layer_input_shape[0],
                               activation_func=activation_func)

        self.layers.append(layer)

        if self.batch_size is not None:
            num_multiplications = np.product(nin) * n_hid + nin[0] * nin[1] * n_hid**2
        else:
            num_multiplications = np.product(nin[1:]) * n_hid + nin[1] * n_hid**2
        _printOps(num_multiplications)
        print '---'
        self.TotalForwardPassCost += num_multiplications

    ############################################################################################################

    def addTiedAutoencoderChain(self,
                                n_layers=None,
                                force_no_dropout=False,
                                activation_func='tanh',
                                input_noise=0.3,
                                tie_W=True):
        """
        Creates connected layers to invert Perceptron layers. Input is assumed to come from the first layer.

        Parameters
        ----------

        n_layers: int
           Number of layers that will be added/inverted, (input < 0 means all)
        activation_func:      string
           {tanh, relu, sigmoid, abs, linear}
           Activation function
        force_no_dropout:     Bool
          set True for last/output layer
        input_noise:         Bool
          Noise rate that will be applied to the input of the first reconstructor
        tie_W:                Bool
          Whether to share weight of dual layer pairs
        """
        if not n_layers:  # Automatically find number of Layers if not specified
            n_layers = len(self.layers)
        assert 0 < n_layers <= len(self.layers), "Number of Autoencoder layers not possible"

        chain = [self.layers[n_layers - 1]]  # if n_layers = depth(NN), add last layer, if n_layers is smaller add
                                             # <n_layers>th layer (s.t. a MLP remains after the AE bzw. next to it))

        first = True
        for i in xrange(n_layers - 1, -1, - 1):  # Invert layers starting from the deepest layer
            n_outputs = self.layers[i].n_in  # Get n_out and Weights from mirror layer
            W = self.layers[i].W.T if tie_W else None
            n_inputs = chain[-1].output_shape[1]  # Get input from previous layer in chain
            # (the first in chain is the deepest layer in the normal Net)
            batch_size = chain[-1].output_shape[0]
            dropout = self._enable_dropout and not force_no_dropout
            noise = input_noise if first else None
            PLayer = PerceptronLayer(input=chain[-1].output,
                                     n_in=n_inputs,
                                     n_out=n_outputs,
                                     batch_size=batch_size,
                                     enable_dropout=dropout,
                                     activation_func=activation_func,
                                     W=W,
                                     input_noise=noise,
                                     input_layer=chain[-1])
            chain.append(PLayer)
            first = False

        self._autoencoder_chains.extend(chain[1:])  # only keep the newly added Layers

        if not tie_W:
            self.layers += chain[1:]

    ############################################################################################################

    def compileDebugFunctions(self, gradients=True):
        """
        Compiles the debug_functions which return the network activations / output. To use them compile them with
        this function. They by accessible as cnn.debug_functions (normal output), cnn.debug_conv_output,
        cnn.debug_gradients_function (if True).
        """
        if len(self.debug_functions) != 0:
            print "debug functions are not empty"
            return

        for lay in self.layers:
            self.debug_functions.append(theano.function([self._x], lay.output))
            try:  # This is the output before pooling etc.
                self.debug_conv_output.append(theano.function(
                    [self._x],
                    lay.conv_output,
                    on_unused_input='ignore'))
            except:
                pass
        if gradients:
            self.debug_gradients_function = opt.Optimizer(self).compileGradients()

    ############################################################################################################

    def compileOutputFunctions(self,
                               target='nll',
                               use_class_weights=False,
                               use_example_weights=False,
                               use_lazy_labels=False,
                               use_label_prop=False,
                               only_forward=False):
        """
        Compiles the output functions ``get_loss``, ``get_error``, ``class_probabilities`` and defines the
        gradient (which is not compiled)

        Parameters
        ----------

        target: string
         'nll'/'regression', regression has squared error and nll_masked allows training with
          lazy labels; this requires the auxiliary (*aux) masks.
        use_class_weights: Bool
          whether to use class weights for the error
        use_example_weights: Bool
          whether to use example weights for the error
        use_lazy_labels: Bool
          whether to use lazy labels; this requires the auxiliary (*aux) masks
        use_label_prop: Bool
          whether to activate label propagation on unlabelled (-1) examples
        only_forwad: Bool
          This exlcudes the building of the gradient (faster)

        Defined functions:

        (They are accessible as methods of ``MixedConvNN``)

        get_loss: theano-function
          [data, labels(, *aux)] --> [loss, loss_instance]
        get_error: theano-function
          [data, labels(, *aux)] --> [loss, (error,) prediction] no error for regression
        class_probabilities: theano-function
          [data] --> [prediction]
        """
        print "GLOBAL"
        _printOps(self.TotalForwardPassCost)
        self.t_graph = time.time()

        for lay in self.layers:
            if lay.params != []:
                self.params.extend(lay.params[::-1])  # (b, W)

        if self._autoencoder_chains is not []:  # add thos layers, but not to the params
            self.layers.extend(self._autoencoder_chains)

        self.param_count = np.sum([np.prod(p.get_value().shape) for p in self.params])

        print "Total Count of trainable Parameters:", self.param_count
        print "Building Computational Graph took %.3f s" % (self.t_graph - self.t_init)
        pp_cw = "using class_weights" if use_class_weights else "using no class_weights"
        pp_ew = "using example_weights" if use_example_weights else "using no example_weights"
        pp_ll = "using lazy_labels" if use_lazy_labels else "using no lazy_labels"
        pp_lp = "label propagation active" if use_label_prop else "label propagation inactive"
        print "Compiling output functions for %s target:\n\t%s\n \t%s\n \t%s\n \t%s\n" % (
            target, pp_cw, pp_ew, pp_ll, pp_lp)

        if len(self._output_layers) != 0:
            print "Warning: <compileOutputFunctions> only applies to the LAST layer in self.layers \
        (and ignores elements of self._output_layers)"

        layer_last = self.layers[-1]
        self.output_shape = layer_last.output_shape
        if (len(layer_last.output_shape) == 2) or self.n_dim != 3:  #Perceptron layer or any other
            self.n_lab = layer_last.output_shape[1]
        else:
            self.n_lab = layer_last.output_shape[2]
        # Define Target functions
        if target == 'regression':
            n_dim_regression = len(layer_last.output_shape)
            self._y = T.TensorType('float32', (False,) * n_dim_regression, name='y_cnn_regression_targets')()
            self._loss, self._loss_instance = layer_last.squared_distance(self._y)
            ret = [self._loss, T.sqrt(self._loss), layer_last.output]
            self.get_error = theano.function([self._x, self._y], ret)
            self.prediction = theano.function([self._x], layer_last.output)

        elif target == 'nll_mutiple_binary':
            self._y = T.wmatrix('y_nll_mutiple_binary_targets')
            if use_class_weights:
                class_weights = T.TensorType('float32', [False], name='class_weights')()
                self._y_aux.append(class_weights)
            else:
                class_weights = None

            self._loss, self._loss_instance = layer_last.nll_mutiple_binary(self._y, class_weights)

        elif target == 'nll_weak':
            if use_class_weights:
                class_weights = T.TensorType('float32', [False], name='class_weights')()
                self._y_aux.append(class_weights)
            else:
                class_weights = None

            self._loss, self._loss_instance = layer_last.NLL_weak(self._y, class_weights)

        elif target == 'affinity':
            self._y = T.TensorType('int16', (False,) * 5, name='y_cnn_affinity_targets')()
            if use_class_weights:
                class_weights = T.TensorType('float32', [False], name='class_weights')()
                self._y_aux.append(class_weights)
            else:
                class_weights = None

            self._loss, self._loss_instance = layer_last.NLL_affinity(self._y, class_weights)

        elif target == 'malis':
            self._y = T.TensorType('int16', (False,) * 5, name='y_cnn_affinity_targets')()
            self._y_aux.append(T.TensorType('int16', (False,) * 4, name='y_cnn_seg_gt')())
            if use_class_weights:
                class_weights = T.TensorType('float32', [False], name='class_weights')()
                self._y_aux.append(class_weights)
            else:
                class_weights = None

            ret = layer_last.NLL_Malis(self._y, self._y_aux[0])
            self._loss = ret[0]
            self._loss_instance = ret[0]
            self.malis_stats = theano.function([self._x, self._y, self._y_aux[0]], ret)

        elif target == 'nll':
            if use_lazy_labels:
                if not (isinstance(layer_last, ConvLayer2d) or isinstance(layer_last, ConvLayer3d)):
                    raise ValueError("Cannot use lazy labels for Percptron layer")

                mask1 = T.TensorType('int16', [False, False], name='mask_class_labeled')()
                self._y_aux.append(mask1)
                mask2 = T.TensorType('int16', [False, False], name='mask_class_not_present')()
                self._y_aux.append(mask2)
            else:
                mask1, mask2 = None, None

            if use_class_weights:
                class_weights = T.TensorType('float32', [False], name='class_weights')()
                self._y_aux.append(class_weights)
            else:
                class_weights = None

            if use_example_weights:
                example_weights = T.TensorType('float32', (False,) * (self._x.ndim - 1), name='example_weights')()
                self._y_aux.append(example_weights)
            else:
                example_weights = None

            if use_label_prop:
                label_prop_thresh = T.fscalar('label_prop_thresh')
                self._y_aux.append(label_prop_thresh)
            else:
                label_prop_thresh = None

            if use_lazy_labels:
                self._loss, self._loss_instance = layer_last.NLL(
                    self._y,
                    class_weights,
                    example_weights,
                    mask_class_labeled=mask1,
                    mask_class_not_present=mask2,
                    label_prop_thresh=label_prop_thresh)
            else:
                self._loss, self._loss_instance = layer_last.NLL(
                    self._y,
                    class_weights,
                    example_weights,
                    label_prop_thresh=label_prop_thresh)

                # aux is possibly [mask_class_labeled, mask_class_not_present, class_weights, label_prop_thresh]

                # For all targets except regression the predictions / accuracy
        if target != 'regression':
            ret = [self._loss, layer_last.errors(self._y), layer_last.class_prediction]
            if target == 'nll_mutiple_binary':
                ret = [self._loss, layer_last.errors_no_tn(self._y), layer_last.class_prediction]

            self.get_error = theano.function([self._x, self._y] + self._y_aux, ret)
            self.class_probabilities = theano.function([self._x], layer_last.class_probabilities)

        # create a list of symbolic gradients for all model parameters
        if not only_forward:
            self._gradients = T.grad(self._loss, self.params, disconnected_inputs="warn")
            self.get_loss = opt.Optimizer(self).get_loss

        if isinstance(layer_last, (ConvLayer2d, ConvLayer3d, AffinityLayer3d)):
            try:
                self.output_shape = layer_last.prob_shape
            except:
                pass

            self.output_strides = map(np.prod, zip(*self.poolings))
            if self.mfp_strides is not None:
                self.output_strides = np.divide(self.output_strides, self.mfp_strides)

        self.t_out = time.time()
        print " Compiling done  - in %.3f s!" % (self.t_out - self.t_graph)
        print '-' * 60
        print '-' * 60

    ############################################################################################################

    def resetMomenta(self):
        """Resets the trailing average of the gradient to sole current gradient"""
        print "CNN: resetting momenta"
        print '\t'.join([str(len(x)) for x in (self.params, self._last_grads)])
        for para, lg in zip(self.params, self._last_grads):
            sp = para.get_value().shape
            lg.set_value(np.zeros(sp, dtype=theano.config.floatX), borrow=0)

        try:
            for para, rp in zip(self.params, self._RPROP_LRs, ):
                sp = para.get_value().shape
                rp.set_value(1e-3 * np.ones(sp, dtype=theano.config.floatX), borrow=0)
        except:
            pass

    def randomizeWeights(self, reset_momenta=True):
        """Resets weights to random values (calls randomize_weights() on each layer)"""
        print "CNN: Randomizing weights"
        for lay in self.layers + self._output_layers:
            lay.randomizeWeights()
        if reset_momenta:
            self.resetMomenta()

    ############################################################################################################
    ### Controlling Training ###################################################################################
    ############################################################################################################

    def trainingStep(self, *args, **kwargs):
        """
        Perform one optimiser iteration.
        Optimizers can be chosen by the kwarg ``mode``. They are complied on demand (which may take a while) and cached

        **Signature**: cnn.trainingStep(data, label(, *aux)(,**kwargs))

        Parameters
        ----------

        data: float32 array
          input [bs, ch (, x, y)] or [bs, z, ch, x, y]
        labels: int16 array
          [bs,((z,)y,x)] if output is not flattened
        aux: int16 arrays
          (optional) auxiliary weights/masks/etc. Should be unpacked list
        kwargs:
          * mode: string
              ['SGD']: (default) Good if data set is big and redundant

              'RPROP': which does neither uses a fix learning rate nor the momentum-value.
              It is faster than SGD if you do full-batch Training and use NO dropout.
              Any source of noise leads to failure of convergence (at all).

              'CG': Good generalisation but requires large batches. Returns current loss always

              'LBFGS': http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.fmin_l_bfgs_b.html


           * update_loss: Bool
               determine current loss *after* update step (e.g. needed for queue, but ``get_loss`` can also be\
               called explicitly)

        Returns
        -------
        loss: float32
          loss (nll or squared error)
        loss_instance: float32 array
          loss for individual batch examples/pixels
        time_per_step: float
          Time spent on the GPU per step
        """
        mode = kwargs.get('mode', 'SGD')
        param_var = None
        t0 = time.time()
        # Check if auxiliary arguments are ok
        if len(args) != (len(self._y_aux) + 2):
            raise ValueError("The number of auxiliary arguments for the NLL is not matching the compiled signature: "
                             "%s. Got %i auxiliary args." % (self._y_aux, len(args) - 2))

        if mode == 'SGD':
            if not hasattr(self, 'SGD'):
                self.SGD = opt.compileSGD(self._SGD_params, self)

            loss, loss_instance = self.SGD(*args)
            if kwargs.get('update_loss', False):
                loss, loss_instance = self.get_loss(*args)

        elif mode == 'RPROP':
            if not hasattr(self, 'RPROP'):
                self.RPROP = opt.compileRPROP(self._RPROP_params, self)

            loss, loss_instance = self.RPROP(*args)
            if kwargs.get('update_loss', False):
                loss, loss_instance = self.get_loss(*args)

        elif mode == 'CG':
            if not hasattr(self, 'CG'):
                self.CG = opt.compileCG(self._CG_params, self)

            loss, loss_instance = self.CG(*args)  # this already is updated loss

        elif mode == 'LBFGS':
            if not hasattr(self, 'LBFGS'):
                self.LBFGS = opt.compileLBFGS(self._LBFGS_params, self)

            loss = self.LBFGS(*args)  # this already is updated loss
            loss_instance = loss

        elif mode == 'Adam':
            if not hasattr(self, 'Adam'):
                self.Adam = opt.compileAdam(self._Adam_params, self)

            loss, loss_instance = self.Adam(*args)  # this already is updated loss

        else:
            print "No mode %s" % mode
            return 0, 0, 0

        t = (time.time() - t0) + 1e-10  # add some epsilon to ensure > 0
        return np.float32(loss), loss_instance, t  ### TODO remove again

    def setOptimizerParams(self,
                           SGD={},
                           CG={},
                           RPROP={},
                           LBFGS={},
                           Adam={},
                           weight_decay=0.0):
        """
        Initialise optimiser hyper-parameters prior to compilation. SGD, CG and LBFGS this can also be done during
        Training.

        ``weight_decay`` is global to all optimisers and
        is identical to a L2-penalty on the weights with the coefficient given by ``weight_decay``
        """
        if weight_decay == False:
            self.global_weightdecay.set_value(np.float32(0), borrow=False)
        else:
            self.global_weightdecay.set_value(np.float32(weight_decay), borrow=False)

        self.setSGDLR(SGD.get("LR", 0.001))
        self.setSGDMomentum(SGD.get("momentum", 0.9))

        self._RPROP_params = dict(penalty=0.35,
                                  gain=0.2,
                                  beta=0.7,
                                  initial_update_size=1e-4)
        self._RPROP_params.update(RPROP)

        self._CG_params = dict(n_steps=3,
                               alpha=0.35,
                               beta=0.7,
                               max_step=0.02,
                               min_step=8e-5,
                               only_descent=False,
                               show=False)
        self._CG_params.update(CG)

        self._LBFGS_params = dict(maxfun= 40,  # function evaluations
                                  maxiter= 4,  # iterations
                                  m= 10,       # maximum number of variable metric corrections
                                  factr= 1e2,  # factor of machine precision as termination criterion (haha!)
                                  pgtol= 1e-9, # projected gradient tolerance
                                  iprint= -1)  # set to 0 for direct printing of steps
        self._LBFGS_params.update(LBFGS)

        self._Adam_params = {}
        self._Adam_params.update(Adam)

        if hasattr(self, 'SSGD'):
            self.SSGD.updateOptimizerParams(SSGD)

        if hasattr(self, 'CG'):
            self.CG.updateOptimizerParams(CG)

        if hasattr(self, 'LBFGS'):
            self.LBFGS.updateOptimizerParams(LBFGS)

    def setSGDLR(self, value=0.09):
        self.SGD_LR.set_value(np.float32(value), borrow=False)

    def setSGDMomentum(self, value=0.9):
        self.SGD_momentum.set_value(np.float32(value), borrow=False)

    def setWeightDecay(self, value=0.0005):
        self.global_weightdecay.set_value(np.float32(value), borrow=False)

    def setDropoutRates(self, rates):
        """Assumes a vector/list/array as input, first entry <--> first layer (etc.)"""
        for lay, ra, i in zip(self.layers, rates, range(len(rates))):
            try:
                assert 1.0 >= np.float32(ra) >= 0, "Dropout rates must be [0,1]"
                lay.activation_noise.set_value(np.float32(ra))
                #print 'layer',i,'new noise rate:',np.float32(ra*100.0),'%'
            except:
                #print 'set_dropout_rates: Warning: Dropout not enabled in this layer'
                pass

    def getDropoutRates(self):
        """Returns list of dropout rates"""
        rates = []
        for lay in self.layers:
            try:
                rates.append(np.float32(lay.activation_noise.get_value()))
            except:
                pass
                #sys.excepthook(*sys.exc_info())
        return rates

    ############################################################################################################
    ### Utilities ##############################################################################################
    ############################################################################################################

    def _predictDenseTile(self, raw_img, out_arr, offset):
        """
        Parameters
        ----------
        raw_img: np.ndarray
          raw image (ch, x, y) or (z, ch, x, y)to be predicted
          The shape must be cnn.patch_size + cnn.output_strides - 1 (elwise)
        out_arr: np.ndarray
          The shape is cnn.patch_size + cnn.mfp_strides - floor(cnn.offset) - 1 (elwise)
        offsets: array / list
          The cnn offsets (only needed if cnn was initialised without a dimension calculator)

        Returns
        -------
        class_probabilities: np.ndarray
          prediction (n_lab, z, x, y)
          The shape is cnn.patch_size + cnn.mfp_strides - floor(cnn.offset) - 1 (elwise)

        """

        if np.all(np.equal(self.output_strides, 1)):
            if self.n_dim == 2:
                out_arr[:, 0] = self.class_probabilities(raw_img[None])[0]  # (ch,x,y)
            else:
                out_arr[:] = self.class_probabilities(raw_img[None])[0]  # (z,ch,x,y)

        else:
            for x_off in range(self.output_strides[-2]):
                for y_off in range(self.output_strides[-1]):
                    if self.n_dim == 2:
                        cut_img = raw_img[None, :, x_off:x_off + self.patch_size[0], y_off:y_off + self.patch_size[1]]

                        #prob = self.class_probabilities(cut_img)[0]
                        # insert prob(ch, x, y) into out_arr(ch,z,x,y)
                        out_arr[:, 0, x_off::self.output_strides[0], y_off::
                                self.output_strides[1]] = self.class_probabilities(cut_img)[0]

                    elif self.n_dim == 3:
                        for z_off in range(self.output_strides[0]):
                            cut_img = raw_img[None,z_off:z_off + self.patch_size[0], :,
                                              x_off:x_off + self.patch_size[1], y_off:y_off + self.patch_size[2]]

                            #prob = self.class_probabilities(cut_img)[0]
                            out_arr[:, z_off::self.output_strides[0], x_off::self.output_strides[1], y_off::
                                    self.output_strides[2]
                                   ] = self.class_probabilities(cut_img)[0]

        return out_arr

    def predictDense(self,
                     raw_img,
                     show_progress=True,
                     offset=None,
                     as_uint8=False,
                     pad_raw=False):
        """
        Core function that performs the inference

        raw_img : np.ndarray
          raw data in the format (ch, x, y(, z))
        show_progress: Bool
          Whether to print progress state
        offset: 2/3-tuple
          If the cnn has no dimension calculator object, this specifies the cnn offset.
        as_uint8: Bool
          Return class proabilites as uint8 image (scaled between 0 and 255!)
        pad_raw: Bool
          Whether to apply padding (by mirroring) to the raw input image
          in order to get predictions on the full imgae domain.
        """
        # WARNING: this code contains mixed orders of xyz and zxy! The raw_img is swapped later!

        # determine normalisation depending on int or float type
        if raw_img.dtype in [np.int, np.int8, np.int16, np.int32, np.uint32,
                             np.uint, np.uint8, np.uint16, np.uint32, np.uint32]:
            m = 255
        else:
            m = 1

        raw_img = np.ascontiguousarray(raw_img, dtype=np.float32) / m

        time_start = time.time()
        strip_z = False
        if len(raw_img.shape) == 3:
            strip_z = True
            raw_img = raw_img[..., None]  # add singleton z-channel
        if self.dimension_calc is not None:
            offset = np.floor(self.dimension_calc.offset).astype(np.int)
        else:
            assert offset is not None,"If the cnn has not been intialised with a dimension calculator object, you must pass the offset to this function explicitly"

            offset = np.floor(offset).astype(np.int)

        n_lab = self.n_lab
        cnn_out_sh = self.output_shape[2:]  # without batch size and channel/n_lab
        ps = self.patch_size
        strides = self.output_strides

        if self.n_dim == 2:
            cnn_out_sh = np.concatenate([[1, ], cnn_out_sh])
            ps = np.concatenate([[1, ], ps])
            strides = np.concatenate([[1, ], strides])
            offset = np.concatenate([[0, ], offset])

        if pad_raw:
            raw_img = np.pad(raw_img, [(0, 0), (offset[1], offset[1]), (offset[2], offset[2]), (offset[0], offset[0])],
                             mode='symmetric')

        raw_sh = raw_img.shape[1:]  # only spatial, not channels

        tile_sh = np.add(ps, strides) - 1  # zxy
        #prob_sh = np.array([ps[i]+strides[i]-1-2*offset[i] for i in xrange(3)]) # zxy
        prob_sh = np.multiply(cnn_out_sh, strides)
        prob_arr = np.zeros(np.concatenate([[self.n_lab, ], prob_sh]), dtype=np.float32)  # zxy

        pred_sh = np.array([raw_sh[0] - 2 * offset[1], raw_sh[1] - 2 * offset[2], raw_sh[2] - 2 * offset[0]])  # xyz
        if as_uint8:
            predictions = np.zeros(np.concatenate(([n_lab, ], pred_sh)), dtype=np.uint8)  # xyz
        else:
            predictions = np.zeros(np.concatenate(([n_lab, ], pred_sh)), dtype=np.float32)  # xyz

        if self._atleast_single_mfp and not np.all(np.equal(self.output_strides, 1)):
            raise NotImplementedError("If MFP is partially enabled, the dense prediction does not work atm")

        # Calculate number of tiles (in 3d: blocks) that need to be performed
        x_tiles = int(np.ceil(float(pred_sh[0]) / prob_sh[1]))
        y_tiles = int(np.ceil(float(pred_sh[1]) / prob_sh[2]))
        z_tiles = int(np.ceil(float(pred_sh[2]) / prob_sh[0]))
        total_nb_tiles = np.product([x_tiles, y_tiles, z_tiles])
        print "Predicting img", raw_img.shape, "in", total_nb_tiles, "Blocks:", (x_tiles, y_tiles, z_tiles)
        count = 0
        for x_t in range(x_tiles):
            for y_t in range(y_tiles):
                for z_t in range(z_tiles):
                    # For every z_tile a slice of thickness cnn_out_sh[2] is
                    # collected and then collectively written to the output_data
                    raw_tile = raw_img[:, x_t * prob_sh[1]:x_t * prob_sh[1] + tile_sh[1],
                                       y_t * prob_sh[2]:y_t * prob_sh[2] + tile_sh[2],
                                       z_t * prob_sh[0]:z_t * prob_sh[0] + tile_sh[0]]

                    this_is_end_tile = False if np.all(np.equal(raw_tile.shape[1:], np.roll(tile_sh, 2))) else True

                    if this_is_end_tile:  # requires 0-padding
                        right_pad = np.subtract(np.roll(tile_sh, 2), raw_tile.shape[1:])  # (ch,x,y,z)
                        right_pad = np.concatenate(([0, ], right_pad))  # for channel dimension
                        left_pad = np.zeros(raw_tile.ndim, dtype=np.int)
                        pad_with = list(zip(left_pad, right_pad))
                        raw_tile = np.pad(raw_tile, pad_with, mode='constant')

                    if self.n_dim == 2:
                        # slice from raw_tile(ch,x,y,z) --> (ch,x,y)
                        prob_arr = self._predictDenseTile(raw_tile[..., 0], prob_arr, offset)  # returns (ch,z=1,x,y)
                        prob = prob_arr[:, 0, :, :, None]  # (ch,z=1,x,y) -> (ch,x,y,z=1)
                    else:
                        raw_tile = np.transpose(raw_tile, (3, 0, 1, 2))  # (ch,x,y,z) -> (z,ch,x,y)
                        prob_arr = self._predictDenseTile(raw_tile, prob_arr, offset)
                        prob = np.transpose(prob_arr, (0, 2, 3, 1))  # (ch,z,x,y) -> (ch,x,y,z)

                    if this_is_end_tile:  # cut away padded range
                        prob = prob[:, :prob_sh[1] - right_pad[1],
                                    :prob_sh[2] - right_pad[2], :prob_sh[0] - right_pad[3]]

                    if as_uint8:
                        prob *= 255
                        prob = prob.astype(np.uint8)  # maybe not needed...

                    predictions[:, x_t * prob_sh[1]:(x_t + 1) * prob_sh[1], y_t * prob_sh[2]:(y_t + 1) * prob_sh[2],
                                z_t * prob_sh[0]:(z_t + 1) * prob_sh[0]] = prob

                    count += 1
                    if show_progress:
                        dtime = time.time() - time_start
                        progress = count * 100.0 / total_nb_tiles
                        estimate = dtime / progress * 100.
                        if progress <= 100:
                            dtime = pprinttime(dtime)
                            estimate = pprinttime(estimate)
                            sys.stdout.write('\rProgress: %.2f%% in %s; estimate: %s' % (progress, dtime, estimate))
                            sys.stdout.flush()

        sys.stdout.write(' - done\n'.decode("string_escape"))
        sys.stdout.flush()
        print "Inference speed: %.3f MB or MPix /s\n" %\
              (np.product(predictions.shape[1:]) * 1.0 / 1000000 / (time.time() - time_start))

        if strip_z: predictions = predictions[:, :, :, 0]

        return predictions

    def get_activities(self, data):
        n = len(self.layers)
        activities = []
        for layer, dbgf, i in zip(self.layers, self.debug_functions, range(n)):
            activities.append(dbgf(data))
        return activities

    def get_nonpooled_activities(self, data):
        n = len(self.layers)
        activities = []
        for layer, dbgf, i in zip(self.layers, self.debug_conv_output, range(n)):
            activities.append(dbgf(data))
        return activities

    def saveParameters(self, path='CNN.save', layers=None, show=True):
        """Saves parameters to file, that can be loaded by ``loadParameters``"""
        if show:
            print 'Saving params to file'
        f = open(path, 'w')
        if layers is None:
            n_lay = len(self.layers) - len(self._autoencoder_chains)  # exclude the AE chains (they have only shared W)
            layers = self.layers[:n_lay]

        shape_info = []
        for lay in layers:
            shape_info.append(lay.params[0].get_value(borrow=True).shape)

        if show:
            print ' shapes are: ' + str(shape_info)
        cPickle.dump(shape_info, f, protocol=2)

        for lay in layers:
            cPickle.dump(lay.params[0].get_value(borrow=True), f, protocol=2)
            cPickle.dump(lay.params[1].get_value(borrow=True), f, protocol=2)
            if len(lay.params) > 2:  # Recurrent Params
                cPickle.dump(lay.params[2].get_value(borrow=True), f, protocol=2)
                cPickle.dump(lay.params[3].get_value(borrow=True), f, protocol=2)

        cPickle.dump(self.poolings, f, protocol=2)  # list of all pooling factors
        f.close()

    def loadParameters(self, myfile="CNN.save", strict=False, n_layers_to_load=-1):
        """
        Loads parameters from file created by ``saveParameters``. The parameter shapes do not need to fit the CNN
        architecture, they "squeezed" or "padded" to fit.

        Additionally the momenta of the gradients are reset

        Parameters
        ----------

        myfile: string
          Path to file
        strict: bool
          If true, parameter shapes must fit exactly, this the only way to load RNN parameters
        n_layers_to_load: int
          Only the first x layers are initialised if this is not at its default value (-1)
        """
        self.resetMomenta()
        if strict:
            self._loadParametersStrict(myfile)
        else:
            self._loadParametersAdaptive(myfile, n_layers_to_load=n_layers_to_load)

    def _loadParametersAdaptive(self, myfile="CNN.save", n_layers_to_load=-1):
        """
        Load a parameter set which is NOT fully compatible to the current network configuration
        (e.g. different filter sizes, number of filters etc).

        detects if layers already are in correct shape
        """
        print "loading(adaptive) from", myfile
        try:
            f = open(myfile, 'r')
        except:
            print "CNN: ERROR: Cannot load file '", myfile, "'"

        shp = cPickle.load(f)
        print "Shapes of loaded file are:", shp
        print "Shapes of current Net are:", [lay.params[0].get_value(borrow=True).shape for lay in self.layers]
        if n_layers_to_load < 0:
            n_layers_to_load = len(shp)
        print "#Layers(sav) =", len(shp), "loading", n_layers_to_load

        print "#Layers(CNN) =", len(self.layers)

        for it, layer in enumerate(self.layers):
            if it == n_layers_to_load:
                break
            if it < len(shp):
                try:
                    p = cPickle.load(f)
                except:
                    print "Error! " * 7
                    print "LoadParametersAdaptive::ERROR: invalid file, cancelled after", it, "layers were loaded!"
                    e = sys.exc_info()[0]
                    print "<p>Error: %s</p>" % e
                    print "Error! " * 7
            else:
                p = [[[[]]]]
                print "debug missing, might crash now!"
            save_shape = np.shape(p)
            target_shape = layer.params[0].get_value(borrow=True).shape

            #load W
            if save_shape == target_shape:
                layer.params[0].set_value(p, borrow=False)
            elif len(target_shape)==len(save_shape) and len(save_shape)==4:
                #temp param of correct shape, weights with same variance as loaded parameters (mean=0)
                temp = np.float32(np.random.normal(0,0.02,target_shape))
                if (target_shape[0]>save_shape[0]):#need more filters than in save
                    for i in range(0,target_shape[0],save_shape[0]):
                        if target_shape[1]>save_shape[1]:
                            for j in range(0, target_shape[1], save_shape[1]):
                                temp[i:min(target_shape[0], save_shape[0] + i),
                                     j:min(target_shape[1], save_shape[1] + j), :min(target_shape[2], save_shape[2]),
                                     :min(target_shape[3], save_shape[3])
                                    ] = p[:(min(target_shape[0], save_shape[0] + i) - i),
                                          :min(target_shape[1] - j, save_shape[1]),
                                          :min(target_shape[2], save_shape[2]),
                                          :min(target_shape[3], save_shape[3])]
                        else:
                            temp[i:min(target_shape[0], save_shape[0] + i), :target_shape[1],
                                 :min(target_shape[2], save_shape[2]),
                                 :min(target_shape[3], save_shape[3])
                                 ] = p[:(min(target_shape[0], save_shape[0] + i) - i),
                                       :target_shape[1],
                                       :min(target_shape[2], save_shape[2]),
                                       :min(target_shape[3], save_shape[3])]

                else:
                    if target_shape[1] > save_shape[1]:
                        for j in range(0, target_shape[1], save_shape[1]):
                            temp[:min(target_shape[0], save_shape[0]), j:min(target_shape[1], save_shape[1] + j),
                                 :min(target_shape[2], save_shape[2]),
                                 :min(target_shape[3], save_shape[3])
                                ] = p[:min(target_shape[0], save_shape[0]),
                                      :min(target_shape[1] - j, save_shape[1]),
                                      :min(target_shape[2], save_shape[2]),
                                      :min(target_shape[3], save_shape[3])
                                     ] + np.random.rand(min(target_shape[0], save_shape[0]),
                                                        min(target_shape[1], save_shape[1]+j)-j,
                                                        min(target_shape[2], save_shape[2]),
                                                        min(target_shape[3], save_shape[3])
                                                        ) * 1e-4
                    else:
                        mid_offset= 0
                        if target_shape[1] < save_shape[1]:
                            mid_offset = int((save_shape[1] - target_shape[1]) / 2.)

                        temp[:min(target_shape[0],save_shape[0]),
                             :target_shape[1],
                             :min(target_shape[2],save_shape[2]),
                             :min(target_shape[3],save_shape[3])
                             ] = p[:min(target_shape[0], save_shape[0]),
                                   mid_offset:mid_offset + target_shape[1],
                                   :min(target_shape[2], save_shape[2]),
                                   :min(target_shape[3], save_shape[3])]

                layer.params[0].set_value(temp, borrow=False)

            elif len(target_shape) == len(save_shape) and len(save_shape) == 5:
                print "adapting 3D_net_filter..."
                #(64, 3, 32, 3, 3) #n. = 64, depth=32
                #print "fan-in correction factor =",n_params_ratio

                temp = np.float32(np.random.normal(0, np.std(p) + 1e-9, target_shape))  #/6.*n_params_ratio

                nf_start = 0
                nf_end = min(target_shape[0], save_shape[0])

                f_st = max(int((target_shape[1] - save_shape[1]) / 2.), 0)
                f_end = f_st + min(target_shape[1], save_shape[1])

                f_st_ = max(int((save_shape[1] - target_shape[1]) / 2.), 0)
                f_end_ = f_st_ + min(target_shape[1], save_shape[1])

                c_st = 0
                c_end = c_st + min(target_shape[2], save_shape[2])  #

                c_st_ = 0
                c_end_ = c_st_ + min(target_shape[2], save_shape[2])  #

                temp[nf_start:nf_end, f_st:f_end, c_st:c_end, f_st:f_end, f_st:f_end
                    ] = p[nf_start:nf_end, f_st_:f_end_, c_st_:c_end_, f_st_:f_end_, f_st_:f_end_]

                layer.params[0].set_value(temp, borrow=False)

            else:
                print "Load: skipping layer" + str(it + 1) + ("" if len(target_shape) == 4 else
                                                              "-            can't load differently shaped perceptron layers (atm)")
            #load b
            if it < len(shp):
                p = cPickle.load(f)
            else:
                p = [[[[]]]]
                print "debug missing"
            save_shape = np.shape(p)
            target_shape = layer.params[1].get_value(borrow=True).shape
            if target_shape[0] == save_shape[0]:
                layer.params[1].set_value(p, borrow=False)
            elif target_shape[0] < save_shape[0]:
                layer.params[1].set_value(p[:min(target_shape[0], save_shape[0])], borrow=False)
            else:
                temp = np.float32(np.random.uniform(1e-5 + (0.5 if layer.activation_func in ["sigmoid", "relu"] else 0),
                                                    1e-6, target_shape))
                for i in range(0, target_shape[0], save_shape[0]):
                    temp[i:min(target_shape[0], save_shape[0] + i)] = p[:min(target_shape[0] - i, save_shape[0])]
                layer.params[1].set_value(temp, borrow=False)
        f.close()
        print "loading complete"

        #function lacks error-handling
    def _loadParametersStrict(self, myfile="CNN.save"):
        """
        Load a parameter set which is **fully compatible** to
        the current network configuration (FAILS otherwise).
        """
        print "Loading from", myfile
        f = open(myfile, 'r')

        shp = cPickle.load(f)
        print "Shapes are:", shp
        print "#Layers =", len(shp)

        for layer in self.layers:
            p = cPickle.load(f)
            layer.params[0].set_value(p, borrow=False)
            p = cPickle.load(f)
            layer.params[1].set_value(p, borrow=False)
            if len(layer.params) == 4:
                p = cPickle.load(f)
                layer.params[2].set_value(p, borrow=False)
                p = cPickle.load(f)
                layer.params[3].set_value(p, borrow=False)

        f.close()

    def gradstats(self, *args, **kwargs):
        grads = self.debug_gradients_function(*args, **kwargs)
        print("Gradient statistics")
        for g in grads:
            print("shape=%s,\tmean=%f,\tstd=%f" %
                  (g.shape, np.mean(g), np.std(g)))

    def actstats(self, *args, **kwargs):
        acts = self.get_activities(*args)
        print("Activation statistics")
        for a in acts:
            print("shape=%s,\tmean=%f,\tstd=%f" %
                  (a.shape, np.mean(a), np.std(a)))

    def paramstats(self, *args, **kwargs):
        print("Parameters statistics")
        for p in self.params:
            p = p.get_value()
            print("shape=%s,\tmean=%f,\tstd=%f" %
                  (p.shape, np.mean(p), np.std(p)))
